#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import defines
from . import queuing


class Graph(object):
    """
    Graph structure and associated algorithms. Made up by a dictionary of
    sources to destinations and costs (weighted edges).
    Edges are unidirectional by default.
    Costs default to a unit.
    """

    def __init__(self, *args):
        self.edges = dict()
        if len(args) > 0 and isinstance(args[0], list):
            self.add_edges(args[0])

    @classmethod
    def _build_path(cls, prev, src, dst):
        """
        Builds the shortest path given dictionary
        of previous nodes.
        """

        cur, path = dst, []
        while not cur == src:
            path.append(cur)
            if not cur in prev:
                return []
            cur = prev[cur]
        path.append(src)
        path.reverse()
        return path

    def add_edges(self, edges):
        for edge in edges:
            if len(edge) < 2:
                continue
            src, dst = edge[0], edge[1]
            cost = edge[2] if len(edge) > 2 and isinstance(edge[2], int) else 1
            bidirectional = (
                edge[3] if len(edge) > 3 and isinstance(edge[3], bool) else False
            )
            self.add_edge(src, dst, cost=cost, bidirectional=bidirectional)

    def add_edge(self, src, dst, cost=1, bidirectional=False):
        if not src in self.edges:
            self.edges[src] = []
        self.edges[src].append((dst, cost))

        if bidirectional:
            self.add_edge(dst, src, cost=cost, bidirectional=False)

    def dijkstra(self, src, dst):
        """
        Dijkstra's algorithm with priority queue implementation
        using the Appier's memory queue.

        :type src: Object
        :param src: The initial node from which a node path to
        the destination should be found.
        :type dst: Object
        :param dst: The destination node to find the path to.
        :see: https://en.wikipedia.org/wiki/Dijkstra%27s_algorithm
        """

        if src == dst:
            return [src], 0

        cls = self.__class__
        dist, prev = dict(), dict()
        dist[src] = 0

        queue = queuing.MemoryQueue()
        queue.push(src, priority=0)

        while queue.length() > 0:
            (_, _, top) = queue.pop(full=True)
            dist[top] = dist[top] if top in dist else defines.INFINITY

            edges = self.edges[top] if top in self.edges else []
            for nxt, cost in edges:
                dist[nxt] = dist[nxt] if nxt in dist else defines.INFINITY

                alt = dist[top] + cost
                if alt < dist[nxt]:
                    dist[nxt] = alt
                    prev[nxt] = top
                    queue.push(nxt, priority=dist[nxt])

        path = cls._build_path(prev, src, dst)
        cost = dist[dst] if dst in dist else defines.INFINITY
        return path, cost



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import pickle
import traceback
import threading

from . import config
from . import legacy
from . import redisdb
from . import component
from . import exceptions


class Bus(component.Component):
    def __init__(self, name="bus", owner=None, *args, **kwargs):
        component.Component.__init__(self, name=name, owner=owner, *args, **kwargs)
        load = kwargs.pop("load", True)
        if load:
            self.load(*args, **kwargs)

    @classmethod
    def new(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def bind(self, name, method):
        raise exceptions.NotImplementedError()

    def unbind(self, name, method=None):
        raise exceptions.NotImplementedError()

    def trigger(self, name, *args, **kwargs):
        raise exceptions.NotImplementedError()


class MemoryBus(Bus):
    def __init__(self, name="memory", owner=None, *args, **kwargs):
        Bus.__init__(self, name=name, owner=owner, *args, **kwargs)

    def bind(self, name, method):
        methods = self._events.get(name, [])
        methods.append(method)
        self._events[name] = methods

    def unbind(self, name, method=None):
        methods = self._events.get(name, [])
        if method:
            methods.remove(method)
        else:
            del methods[:]

    def trigger(self, name, *args, **kwargs):
        methods = self._events.get(name, [])
        for method in methods:
            method(*args, **kwargs)

    def _load(self, *args, **kwargs):
        Bus._load(self, *args, **kwargs)
        self._events = dict()

    def _unload(self, *args, **kwargs):
        Bus._unload(self, *args, **kwargs)
        self._events = None

    def _get_state(self):
        state = Bus._get_state(self)
        state.update(_events=self._events)
        return state

    def _set_state(self, state):
        Bus._set_state(self, state)
        _events = state.get("_events", {})
        for name, methods in legacy.iteritems(_events):
            for method in methods:
                self.bind(name, method)


class RedisBus(Bus):
    SERIALIZER = pickle
    """ The serializer to be used for the values contained in
    the bus (used on top of the class) """

    GLOBAL_CHANNEL = "global"
    """ The name of the global channel to which all of the agents
    should be subscribed, for global communication """

    def __init__(self, name="redis", owner=None, *args, **kwargs):
        Bus.__init__(self, name=name, owner=owner, *args, **kwargs)
        self._delay = kwargs.pop("delay", True)

    def bind(self, name, method):
        methods = self._events.get(name, [])
        methods.append(method)
        self._events[name] = methods
        channel = self._to_channel(name)
        self._pubsub.subscribe(channel)

    def unbind(self, name, method=None):
        methods = self._events.get(name, [])
        if method:
            methods.remove(method)
        else:
            del methods[:]
        channel = self._to_channel(name)
        self._pubsub.unsubscribe(channel)

    def trigger(self, name, *args, **kwargs):
        channel = self._to_channel(name)
        data = self._serializer.dumps(dict(args=args, kwargs=kwargs))
        self._redis.publish(channel, data)

    def _load(self, *args, **kwargs):
        Bus._load(self, *args, **kwargs)
        self._name = config.conf("BUS_NAME", self.owner.name_i)
        self._name = config.conf("BUS_SCOPE", self._name)
        self._name = kwargs.pop("name", self._name)
        self._serializer = kwargs.pop("serializer", self.__class__.SERIALIZER)
        self._global_channel = kwargs.pop(
            "global_channel", self.__class__.GLOBAL_CHANNEL
        )
        self._events = dict()
        self._open()

    def _unload(self, *args, **kwargs):
        Bus._unload(self, *args, **kwargs)
        self._events = None
        self._close()

    def _get_state(self):
        state = Bus._get_state(self)
        state.update(_events=self._events)
        return state

    def _set_state(self, state):
        Bus._set_state(self, state)
        _events = state.get("_events", {})
        for name, methods in legacy.iteritems(_events):
            for method in methods:
                self.bind(name, method)

    def _open(self):
        cls = self.__class__
        channel = self._to_channel(self._global_channel)
        self._redis = redisdb.get_connection()
        self._redis.ping()
        self._pubsub = self._redis.pubsub()
        self._pubsub.subscribe(channel)
        self._listener = RedisListener(self)
        self._listener.start()

    def _close(self):
        self._pubsub.unsubscribe()
        self._listener.join()
        self._redis = None
        self._pubsub = None
        self._listener = None

    def _loop(self, safe=True):
        for item in self._pubsub.listen():
            try:
                if not self.loaded:
                    break
                self._tick(item, safe=safe)
            except Exception as exception:
                self.logger.critical("Unhandled Redis loop exception raised")
                self.logger.error(exception)
                lines = traceback.format_exc().splitlines()
                for line in lines:
                    self.logger.warning(line)

    def _tick(self, item, safe=True):
        channel = item.get("channel", None)
        channel = legacy.str(channel)
        type = item.get("type", None)
        data = item.get("data", None)
        if not type in ("message",):
            return
        if ":" in channel:
            _prefix, name = channel.split(":", 1)
        else:
            name = channel
        data = self._serializer.loads(data)
        methods = self._events.get(name, []) if self._events else []
        for method in methods:
            if safe:
                self.owner.schedule(
                    method,
                    args=data["args"],
                    kwargs=data["kwargs"],
                    timeout=-1,
                    safe=True,
                )
            else:
                method(*data["args"], **data["kwargs"])

    def _to_channel(self, name):
        return self._name + ":" + name


class RedisListener(threading.Thread):
    def __init__(self, bus):
        threading.Thread.__init__(self, name="RedisListener")
        self.daemon = True
        self._bus = bus

    def run(self):
        self._bus._loop()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import json
import uuid

from . import common
from . import legacy


class AppierException(Exception):
    """
    Top level exception to be used as the root of
    all the exceptions to be raised by the appier infra-
    structure. Should be compatible with HTTP status
    codes for proper HTTP serialization.
    """

    message = None
    """ The message value stored to describe the
    current exception value """

    code = None
    """ The internal error code to be used in objective
    serialization of the error (eg: HTTP) """

    headers = None
    """ Optional list of MIME compliant headers to be sent
    in a client/server paradigm """

    meta = None
    """ The meta information associated with the error
    that should be considered private in context, this
    should be structured data ready to be serializable """

    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args)
        self.name = self._name()
        self.message = kwargs.get("message", self.name)
        self.code = kwargs.get("code", 500)
        self.headers = kwargs.get("headers", None)
        self.meta = kwargs.get("meta", None)
        self.args = args
        self.kwargs = kwargs
        self._uid = None

    def __str__(self):
        if legacy.PYTHON_3:
            return self.__unicode__()
        is_unicode = legacy.is_unicode(self.message)
        if is_unicode:
            return self.message.encode("utf-8")
        return self.message

    def __unicode__(self):
        is_unicode = legacy.is_unicode(self.message)
        if not is_unicode:
            return self.message.decode("utf-8")
        return self.message

    def get_meta(self, name, default=None):
        if not self.meta:
            return default
        return self.meta.get(name, default)

    def set_meta(self, name, value):
        if not self.meta:
            self.meta = {}
        self.meta[name] = value

    def del_meta(self, name):
        if not self.meta:
            return
        if not name in self.meta:
            return
        del self.meta[name]

    @property
    def uid(self):
        if self._uid:
            return self._uid
        self._uid = uuid.uuid4()
        return self._uid

    def _name(self):
        cls = self.__class__
        return common.util().camel_to_readable(cls.__name__)


class OperationalError(AppierException):
    """
    Error raised for a runtime error and as a result
    of an operational routine that failed.
    This should not be used for coherent development
    bugs, that are raised continuously.
    """

    pass


class SecurityError(AppierException):
    """
    Error used to indicate security problems that may
    arise during the execution (runtime) of an appier
    application. This error should not be used to notify
    development related problems.
    """

    pass


class AssertionError(OperationalError):
    """
    Error raised for failure to meet any pre-condition or
    assertion for a certain data set.
    """

    def __init__(self, *args, **kwargs):
        kwargs["message"] = kwargs.get("message", "Assertion of data failed")
        kwargs["code"] = kwargs.get("code", None)
        OperationalError.__init__(self, *args, **kwargs)


class ValidationError(OperationalError):
    """
    Error raised when a validation on the model fails
    the error should associate a name in the model with
    a message describing the validation failure.
    """

    errors = None
    """ The map containing an association between the name
    of a field and a list of validation errors for it """

    model = None
    """ The model containing the values in it after the
    process of validation has completed """

    def __init__(self, errors, model, *args, **kwargs):
        kwargs["message"] = kwargs.get("message", "Validation of submitted data failed")
        kwargs["code"] = kwargs.get("code", 400)
        OperationalError.__init__(self, *args, **kwargs)
        self.errors = errors
        self.model = model
        self.set_meta("errors", self.errors)

    def __str__(self):
        if legacy.PYTHON_3:
            return self.__unicode__()
        message = OperationalError.__str__(self)
        extended = common.is_devel()
        if not extended:
            return message
        errors_s = self.errors_s()
        if not errors_s:
            return message
        if self.model:
            message += " for model '%s' with id '%s'" % (
                self.model.__class__._name(),
                self.model._id if hasattr(self.model, "_id") else "unset",
            )
        errors_s = errors_s.encode("utf-8")
        message += " (" + errors_s + ")"
        return message

    def __unicode__(self):
        message = OperationalError.__unicode__(self)
        extended = common.is_devel()
        if not extended:
            return message
        errors_s = self.errors_s()
        if not errors_s:
            return message
        if self.model:
            message += " for model '%s' with id '%s'" % (
                self.model.__class__._name(),
                self.model._id if hasattr(self.model, "_id") else "unset",
            )
        message += " (" + errors_s + ")"
        return message

    def errors_s(self, encoding="utf-8"):
        if not self.errors:
            return ""
        buffer = []
        is_first = True
        for name, errors in legacy.iteritems(self.errors):
            for error in errors:
                is_bytes = legacy.is_bytes(error)
                if is_bytes:
                    error = error.decode(encoding)
                if is_first:
                    is_first = False
                else:
                    buffer.append(", ")
                buffer.append(name)
                buffer.append(" => ")
                buffer.append(error)
        return legacy.u("").join(buffer)


class NotFoundError(OperationalError):
    """
    Error originated from an operation that was not able
    to be performed because it was not able to found the
    requested entity/value as defined by specification.
    """

    def __init__(self, *args, **kwargs):
        kwargs["code"] = kwargs.get("code", 404)
        OperationalError.__init__(self, *args, **kwargs)


class NotImplementedError(OperationalError):
    """
    Error to be raised when a certain feature or route is not
    yet implemented or is not meant to be implemented at the
    defined abstraction level.
    """

    def __init__(self, *args, **kwargs):
        kwargs["code"] = kwargs.get("code", 501)
        OperationalError.__init__(self, *args, **kwargs)


class BaseInternalError(RuntimeError):
    """
    The base (internal) error class from which all the
    internal error classes should inherit, contains basic
    functionality to be inherited by all the internal "errors".
    """

    message = None
    """ The message value stored to describe the
    current error, this should be a valid string """

    meta = None
    """ The meta information associated with the error
    that should be considered private in context, this
    should be structured data ready to be serializable """

    def __init__(self, message, meta=None):
        RuntimeError.__init__(self, message)
        self.message = message
        self.meta = meta

    def __str__(self):
        if legacy.PYTHON_3:
            return self.__unicode__()
        is_unicode = legacy.is_unicode(self.message)
        if is_unicode:
            return self.message.encode("utf-8")
        return self.message

    def __unicode__(self):
        is_unicode = legacy.is_unicode(self.message)
        if not is_unicode:
            return self.message.decode("utf-8")
        return self.message

    def get_meta(self, name, default=None):
        if not self.meta:
            return default
        return self.meta.get(name, default)

    def set_meta(self, name, value):
        if not self.meta:
            self.meta = {}
        self.meta[name] = value

    def del_meta(self, name):
        if not self.meta:
            return
        if not name in self.meta:
            return
        del self.meta[name]


class ValidationInternalError(BaseInternalError):
    """
    Error raised when a validation on the model fails
    the error should associate a name in the model with
    a message describing the validation failure.
    """

    name = None
    """ The name of the attribute that failed
    the validation, for latter reference """

    def __init__(self, name, message, meta=None):
        BaseInternalError.__init__(self, message, meta=meta)
        self.name = name


class ValidationMultipleError(ValidationInternalError):
    """
    Exception/error considered to be equivalent to the
    validation internal error, with the exception that it
    may handle multiple errors at the same time.
    """

    errors = []
    """ The sequence containing the multiple errors associated
    with the validation multiple error """

    def __init__(self, name=None, message=None, meta=None):
        ValidationInternalError.__init__(self, name, message, meta=meta)
        self.errors = []
        self.set_meta("errors", self.errors)

    def add_error(self, name, message):
        if not self.name:
            self.name = name
        if not self.message:
            self.message = message
        self.errors.append((name, message))

    def add_exception(self, exception):
        self.add_error(exception.name, exception.message)


class HTTPError(BaseInternalError):
    """
    Top level HTTP error raised whenever a bad response
    is received from the server peer. This error is meant
    to be used by the client library.
    """

    error = None
    """ The reference to the original and internal
    HTTP error that is going to be used in the reading
    of the underlying internal buffer """

    _data = None
    """ The underlying/internal data attribute that is
    going to be used to cache the binary contents of the
    error associated with this exception (data) stream """

    def __init__(self, error, code=None, message=None, extended=None, meta=None):
        message = message or "Problem in the HTTP request"
        if extended == None:
            extended = common.is_devel()
        if code:
            message = "[%d] %s" % (code, message)
        if extended:
            data = self.read(error=error)
            try:
                data = data.decode("utf-8")
            except Exception:
                data = legacy.str(data)
            if data:
                message = message + "\n" + data if message else data
        BaseInternalError.__init__(self, message, meta=meta)
        self.code = code
        self.error = error

    def read(self, error=None):
        error = error or self.error
        if not self._data == None:
            return self._data
        self._data = error.read()
        return self._data

    def read_json(self, error=None):
        error = error or self.error
        data = self.read(error=error)
        if legacy.is_bytes(data):
            data = data.decode("utf-8")
        try:
            data_j = json.loads(data)
        except Exception:
            data_j = None
        return data_j


class APIError(BaseInternalError):
    """
    Highest level error for API related problems that may be
    raised from the appier API infra-structure. These kind of
    errors should be encapsulated around proper structures
    """

    def __init__(self, *args, **kwargs):
        message = kwargs.get("message", None)
        meta = kwargs.get("meta", None)
        BaseInternalError.__init__(self, message, meta=meta)


class APIAccessError(APIError):
    """
    General purpose access error exception, to be raised under
    situations where the access to a certain functionalities is
    denied for insufficient permissions/invalid credentials.
    """

    original = None
    """ The reference to the original error/exception that originated
    this API access error, this may be unset in case no concrete
    error has originated this error """

    def __init__(self, *args, **kwargs):
        self.original = kwargs.get("original", None)
        if self.original and hasattr(self.original, "message"):
            kwargs["message"] = self.original.message
        APIError.__init__(self, *args, **kwargs)


class OAuthAccessError(APIAccessError):
    """
    OAuth related problems that typically involve either outdated
    tokens or invalid ones. Triggering this exception should imply
    a revalidation of the current token.
    """

    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import amqp
from . import api
from . import asynchronous
from . import base
from . import bus
from . import cache
from . import component
from . import compress
from . import config
from . import controller
from . import crypt
from . import data
from . import defines
from . import exceptions
from . import execution
from . import export
from . import extra
from . import geo
from . import git
from . import graph
from . import http
from . import legacy
from . import log
from . import meta
from . import mock
from . import model
from . import mongo
from . import observer
from . import part
from . import preferences
from . import queuing
from . import redisdb
from . import request
from . import scheduler
from . import serialize
from . import session
from . import settings
from . import smtp
from . import storage
from . import structures
from . import typesf
from . import util
from . import validation

from .amqp import AMQP
from .api import API, OAuthAPI, OAuth1API, OAuth2API
from .asynchronous import (
    ASYNC_HEADER,
    AsyncManager,
    SimpleManager,
    QueueManager,
    AwaitWrapper,
    CoroutineWrapper,
    AyncgenWrapper,
    await_wrap,
    await_yield,
    ensure_generator,
    is_coroutine,
    is_coroutine_object,
    is_coroutine_native,
    to_coroutine,
    wrap_silent,
    unavailable,
    is_neo,
    Future,
    coroutine,
    wakeup,
    sleep,
    wait,
    notify,
    build_future,
    ensure_async,
    header_a,
    ensure_a,
)
from .base import (
    APP,
    LEVEL,
    NAME,
    VERSION,
    PLATFORM,
    IDENTIFIER_SHORT,
    IDENTIFIER_LONG,
    IDENTIFIER,
    API_VERSION,
    BUFFER_SIZE,
    MAX_LOG_SIZE,
    MAX_LOG_COUNT,
    App,
    APIApp,
    WebApp,
    Template,
    get_app,
    get_name,
    get_base_path,
    get_cache,
    get_preferences,
    get_bus,
    get_request,
    get_session,
    get_model,
    get_controller,
    get_part,
    get_adapter,
    get_manager,
    get_logger,
    get_level,
    is_loaded,
    is_devel,
    is_safe,
    to_locale,
    on_exit,
    build_asgi,
    build_asgi_i,
)
from .bus import Bus, MemoryBus, RedisBus
from .cache import Cache, MemoryCache, FileCache, RedisCache, SerializedCache
from .component import Component
from .compress import Compress
from .config import conf, conf_prefix, conf_suffix, conf_s, conf_r, conf_d, conf_ctx
from .controller import Controller
from .crypt import Cipher, RC4, Spritz
from .data import (
    DataAdapter,
    MongoAdapter,
    TinyAdapter,
    Collection,
    MongoCollection,
    TinyCollection,
)
from .defines import (
    ITERABLES,
    MOBILE_REGEX,
    TABLET_REGEX,
    MOBILE_PREFIX_REGEX,
    BODY_REGEX,
    TAG_REGEX,
    EMAIL_REGEX,
    BROWSER_INFO,
    OS_INFO,
    WINDOWS_LOCALE,
    SLUG_PERMUTATIONS,
)
from .exceptions import (
    AppierException,
    OperationalError,
    SecurityError,
    AssertionError,
    ValidationError,
    NotFoundError,
    NotImplementedError,
    BaseInternalError,
    ValidationInternalError,
    ValidationMultipleError,
    HTTPError,
    APIError,
    APIAccessError,
    OAuthAccessError,
)
from .execution import (
    ExecutionThread,
    background,
    insert_work,
    interval_work,
    seconds_work,
    minutes_work,
    hourly_work,
    daily_work,
    weekly_work,
    monthly_work,
    seconds_eval,
    minutes_eval,
    hourly_eval,
    daily_eval,
    weekly_eval,
    monthly_eval,
)
from .export import ExportManager
from .extra import (
    get_a,
    post_a,
    put_a,
    delete_a,
    patch_a,
    get_w,
    post_w,
    put_w,
    delete_w,
    patch_w,
)
from .geo import GeoResolver
from .git import Git
from .graph import Graph
from .http import file_g, get_f, get, post, put, delete, patch, basic_auth, HTTPResponse
from .log import (
    MemoryHandler,
    BaseFormatter,
    ThreadFormatter,
    DummyLogger,
    reload_format,
    rotating_handler,
    smtp_handler,
    in_signature,
)
from .meta import Ordered, Indexed
from .mock import MockObject, MockResponse, MockApp
from .model import (
    Model,
    LocalModel,
    Field,
    link,
    operation,
    view,
    field,
    type_d,
    is_unset,
)
from .mongo import (
    Mongo,
    MongoAsync,
    MongoEncoder,
    get_connection,
    get_connection_a,
    reset_connection,
    reset_connection_a,
    get_db,
    get_db_a,
    drop_db,
    drop_db_a,
    object_id,
    dumps,
)
from .observer import Observable
from .part import Part
from .preferences import (
    Preferences,
    MemoryPreferences,
    FilePreferences,
    RedisPreferences,
)
from .queuing import Queue, MemoryQueue, MultiprocessQueue, AMQPQueue
from .redisdb import Redis
from .request import CODE_STRINGS, Request, MockRequest
from .scheduler import Scheduler, CronScheduler, SchedulerTask, SchedulerDate, Cron
from .serialize import serialize_csv, serialize_ics, build_encoder
from .session import (
    Session,
    MockSession,
    MemorySession,
    FileSession,
    RedisSession,
    ClientSession,
)
from .settings import DEBUG, USERNAME, PASSWORD
from .smtp import (
    message,
    message_base,
    message_netius,
    smtp_engine,
    multipart,
    plain,
    html,
    application,
    header,
)
from .storage import StorageEngine, BaseEngine, FsEngine
from .structures import OrderedDict, LazyDict, LazyValue, GeneratorFile, lazy_dict, lazy
from .typesf import (
    AbstractType,
    Type,
    File,
    Files,
    ImageFile,
    ImageFiles,
    image,
    images,
    Reference,
    reference,
    References,
    references,
    Encrypted,
    encrypted,
    secure,
)
from .util import (
    to_limit,
    to_find,
    to_sort,
    is_iterable,
    is_mobile,
    is_tablet,
    is_browser,
    is_bot,
    browser_info,
    email_parts,
    email_mime,
    email_name,
    email_base,
    date_to_timestamp,
    obfuscate,
    import_pip,
    ensure_pip,
    install_pip,
    install_pip_s,
    request_json,
    get_context,
    get_object,
    resolve_alias,
    page_types,
    find_types,
    norm_object,
    set_object,
    leafs,
    gather_errors,
    gen_token,
    html_to_text,
    camel_to_underscore,
    camel_to_readable,
    underscore_to_camel,
    underscore_to_readable,
    quote,
    unquote,
    escape,
    unescape,
    count_unescape,
    split_unescape,
    call_safe,
    base_name,
    base_name_m,
    is_content_type,
    parse_content_type,
    parse_cookie,
    parse_multipart,
    decode_params,
    load_form,
    check_login,
    check_user,
    check_token,
    check_tokens,
    ensure_login,
    get_tokens_m,
    to_tokens_m,
    dict_merge,
    deprecated,
    cached,
    private,
    ensure,
    delayed,
    route,
    error_handler,
    exception_handler,
    before_request,
    after_request,
    is_detached,
    sanitize,
    verify,
    verify_equal,
    verify_not_equal,
    verify_type,
    verify_many,
    execute,
    ctx_locale,
    ctx_request,
    FileTuple,
    BaseThread,
    JSONEncoder,
)
from .validation import (
    validate,
    validate_b,
    validate_e,
    safe,
    eq,
    gt,
    gte,
    lt,
    lte,
    not_null,
    not_empty,
    not_false,
    is_in,
    is_upper,
    is_lower,
    is_simple,
    is_email,
    is_url,
    is_regex,
    field_eq,
    field_gt,
    field_gte,
    field_lt,
    field_lte,
    string_gt,
    string_lt,
    string_eq,
    equals,
    not_past,
    not_duplicate,
    all_different,
    no_self,
)

from .amqp import get_connection as get_amqp
from .amqp import properties as properties_amqp
from .mongo import get_connection as get_mongo
from .mongo import get_db as get_mongo_db
from .mongo import drop_db as drop_mongo_db
from .mongo import object_id as object_id_mongo
from .mongo import dumps as dumps_mongo
from .mongo import serialize as serialize_mongo
from .mongo import directions as directions_mongo
from .redisdb import get_connection as get_redis
from .redisdb import dumps as dumps_redis

HTTPError = exceptions.HTTPError



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import sys
import calendar
import datetime
import inspect
import functools
import itertools
import contextlib
import collections

import urllib  # @UnusedImport

ArgSpec = collections.namedtuple("ArgSpec", ["args", "varargs", "keywords", "defaults"])


@contextlib.contextmanager
def ctx_absolute():
    root = sys.path.pop(0)
    try:
        yield
    finally:
        sys.path.insert(0, root)


with ctx_absolute():
    try:
        import urllib2
    except ImportError:
        urllib2 = None

with ctx_absolute():
    try:
        import httplib
    except ImportError:
        httplib = None

with ctx_absolute():
    try:
        import http
    except ImportError:
        http = None

with ctx_absolute():
    try:
        import types
    except ImportError:
        types = None

with ctx_absolute():
    try:
        import urllib.error
    except ImportError:
        pass

with ctx_absolute():
    try:
        import urllib.request
    except ImportError:
        pass

with ctx_absolute():
    try:
        import http.client
    except ImportError:
        pass

with ctx_absolute():
    try:
        import importlib.util
    except ImportError:
        pass

try:
    import HTMLParser
except ImportError:
    import html.parser

    HTMLParser = html.parser

try:
    import cPickle
except ImportError:
    import pickle

    cPickle = pickle

try:
    import imp
except ImportError:
    import importlib

    imp = importlib

try:
    import importlib
except ImportError:
    import imp

    importlib = imp

try:
    import cStringIO
except ImportError:
    import io

    cStringIO = io

try:
    import StringIO as _StringIO
except ImportError:
    import io

    _StringIO = io

try:
    import urlparse as _urlparse
except ImportError:
    import urllib.parse

    _urlparse = urllib.parse

PYTHON_3 = sys.version_info[0] >= 3
""" Global variable that defines if the current Python
interpreter is at least Python 3 compliant, this is used
to take some of the conversion decision for runtime """

PYTHON_33 = sys.version_info[0] >= 3 and sys.version_info[1] >= 3
""" Global variable that defines if the current Python
interpreter is at least Python 3.3 compliant """

PYTHON_35 = sys.version_info[0] >= 3 and sys.version_info[1] >= 5
""" Global variable that defines if the current Python
interpreter is at least Python 3.5 compliant """

PYTHON_36 = sys.version_info[0] >= 3 and sys.version_info[1] >= 6
""" Global variable that defines if the current Python
interpreter is at least Python 3.6 compliant """

PYTHON_39 = sys.version_info[0] >= 3 and sys.version_info[1] >= 9
""" Global variable that defines if the current Python
interpreter is at least Python 3.9 compliant """

PYTHON_ASYNC = PYTHON_35
""" Global variable that defines if the current Python
interpreter support the async/await syntax responsible
for the easy to use async methods """

PYTHON_ASYNC_GEN = PYTHON_36
""" Global variable that defines if the current Python
interpreter support the async/await generator syntax
responsible for the async generator methods """

PYTHON_V = int("".join([str(v) for v in sys.version_info[:3]]))
""" The Python version integer describing the version of
a the interpreter as a set of three integer digits """

if PYTHON_3:
    LONG = int
else:
    LONG = long  # @UndefinedVariable

if PYTHON_3:
    BYTES = bytes
else:
    BYTES = str  # @UndefinedVariable

if PYTHON_3:
    UNICODE = str
else:
    UNICODE = unicode  # @UndefinedVariable

if PYTHON_3:
    OLD_UNICODE = None
else:
    OLD_UNICODE = unicode  # @UndefinedVariable

if PYTHON_3:
    STRINGS = (str,)
else:
    STRINGS = (str, unicode)  # @UndefinedVariable

if PYTHON_3:
    ALL_STRINGS = (bytes, str)
else:
    ALL_STRINGS = (bytes, str, unicode)  # @UndefinedVariable

if PYTHON_3:
    INTEGERS = (int,)
else:
    INTEGERS = (int, long)  # @UndefinedVariable

# saves a series of global symbols that are going to be
# used latter for some of the legacy operations
_ord = ord
_chr = chr
_str = str
_bytes = bytes
_range = range

try:
    _xrange = xrange  # @UndefinedVariable
except Exception:
    _xrange = None

if PYTHON_3:
    Request = urllib.request.Request
else:
    Request = urllib2.Request

if PYTHON_3:
    HTTPHandler = urllib.request.HTTPHandler
else:
    HTTPHandler = urllib2.HTTPHandler

if PYTHON_3:
    HTTPError = urllib.error.HTTPError
else:
    HTTPError = urllib2.HTTPError

if PYTHON_3:
    HTTPConnection = http.client.HTTPConnection  # @UndefinedVariable
else:
    HTTPConnection = httplib.HTTPConnection

if PYTHON_3:
    HTTPSConnection = http.client.HTTPSConnection  # @UndefinedVariable
else:
    HTTPSConnection = httplib.HTTPSConnection

try:
    _execfile = execfile  # @UndefinedVariable
except Exception:
    _execfile = None

try:
    _reduce = reduce  # @UndefinedVariable
except Exception:
    _reduce = None

try:
    _reload = reload  # @UndefinedVariable
except Exception:
    _reload = None

try:
    _unichr = unichr  # @UndefinedVariable
except Exception:
    _unichr = None


def with_meta(meta, *bases):
    return meta("Class", bases, {})


def eager(iterable):
    if PYTHON_3:
        return list(iterable)
    return iterable


def iteritems(associative):
    if PYTHON_3:
        return associative.items()
    return associative.iteritems()


def iterkeys(associative):
    if PYTHON_3:
        return associative.keys()
    return associative.iterkeys()


def itervalues(associative):
    if PYTHON_3:
        return associative.values()
    return associative.itervalues()


def items(associative):
    if PYTHON_3:
        return eager(associative.items())
    return associative.items()


def keys(associative):
    if PYTHON_3:
        return eager(associative.keys())
    return associative.keys()


def values(associative):
    if PYTHON_3:
        return eager(associative.values())
    return associative.values()


def xrange(start, stop=None, step=1):
    if PYTHON_3:
        return _range(start, stop, step) if stop else _range(start)
    return _xrange(start, stop, step) if stop else _range(start)


def range(start, stop=None, step=None):
    if PYTHON_3:
        return eager(_range(start, stop, step)) if stop else eager(_range(start))
    return _range(start, stop, step) if stop else _range(start)


def ord(value):
    if PYTHON_3 and type(value) == int:
        return value
    return _ord(value)


def chr(value):
    if PYTHON_3:
        return _bytes([value])
    if type(value) in INTEGERS:
        return _chr(value)
    return value


def chri(value):
    if PYTHON_3:
        return value
    if type(value) in INTEGERS:
        return _chr(value)
    return value


def bytes(value, encoding="latin-1", errors="strict", force=False):
    if not PYTHON_3 and not force:
        return value
    if value == None:
        return value
    if type(value) == _bytes:
        return value
    return value.encode(encoding, errors)


def str(value, encoding="latin-1", errors="strict", force=False):
    if not PYTHON_3 and not force:
        return value
    if value == None:
        return value
    if type(value) in STRINGS:
        return value
    return value.decode(encoding, errors)


def u(value, encoding="utf-8", errors="strict", force=False):
    if PYTHON_3 and not force:
        return value
    if value == None:
        return value
    if type(value) == UNICODE:
        return value
    return value.decode(encoding, errors)


def ascii(value, encoding="utf-8", errors="replace"):
    if is_bytes(value):
        value = value.decode(encoding, errors)
    else:
        value = UNICODE(value)
    value = value.encode("ascii", errors)
    value = str(value)
    return value


def orderable(value):
    if not PYTHON_3:
        return value
    return Orderable(value)


def is_str(value):
    return type(value) == _str


def is_unicode(value):
    if PYTHON_3:
        return type(value) == _str
    else:
        return type(value) == unicode  # @UndefinedVariable


def is_bytes(value):
    if PYTHON_3:
        return type(value) == _bytes
    else:
        return type(value) == _str  # @UndefinedVariable


def is_string(value, all=False):
    target = ALL_STRINGS if all else STRINGS
    return type(value) in target


def is_generator(value):
    if inspect.isgenerator(value):
        return True
    if type(value) in (itertools.chain,):
        return True
    if hasattr(value, "_is_generator"):
        return True
    return False


def is_async_generator(value):
    if not hasattr(inspect, "isasyncgen"):
        return False
    return inspect.isasyncgen(value)


def is_unittest(name="unittest"):
    current_stack = inspect.stack()
    for stack_frame in current_stack:
        for program_line in stack_frame[4]:
            is_unittest = not name in program_line
            if is_unittest:
                continue
            return True
    return False


def execfile(path, global_vars, local_vars=None, encoding="utf-8"):
    if local_vars == None:
        local_vars = global_vars
    if not PYTHON_3:
        return _execfile(path, global_vars, local_vars)
    file = open(path, "rb")
    try:
        data = file.read()
    finally:
        file.close()
    data = data.decode(encoding)
    code = compile(data, path, "exec")
    exec(code, global_vars, local_vars)  # @UndefinedVariable


def walk(path, visit, arg):
    for root, dirs, _files in os.walk(path):
        names = os.listdir(root)
        visit(arg, root, names)
        for dir in list(dirs):
            exists = dir in names
            not exists and dirs.remove(dir)


def getargspec(func):
    has_full = hasattr(inspect, "getfullargspec")
    if has_full:
        return ArgSpec(*inspect.getfullargspec(func)[:4])
    else:
        return inspect.getargspec(func)


def has_module(name):
    if PYTHON_3:
        try:
            spec = importlib.util.find_spec(name)
        except ImportError:
            return False
        if spec == None:
            return False
        return True
    try:
        file, _path, _description = imp.find_module(name)
    except ImportError:
        return False
    if file:
        file.close()
    return True


def new_module(name):
    if hasattr(types, "ModuleType"):
        return types.ModuleType(name)
    if hasattr(imp, "new_module"):
        return imp.new_module(name)
    raise ValueError("No module build method available")


def reduce(*args, **kwargs):
    if PYTHON_3:
        return functools.reduce(*args, **kwargs)
    return _reduce(*args, **kwargs)


def reload(*args, **kwargs):
    if PYTHON_3:
        return importlib.reload(*args, **kwargs)
    return _reload(*args, **kwargs)


def unichr(*args, **kwargs):
    if PYTHON_3:
        return _chr(*args, **kwargs)
    return _unichr(*args, **kwargs)


def urlopen(*args, **kwargs):
    if PYTHON_3:
        return urllib.request.urlopen(*args, **kwargs)
    else:
        return urllib2.urlopen(*args, **kwargs)  # @UndefinedVariable


def build_opener(*args, **kwargs):
    if PYTHON_3:
        return urllib.request.build_opener(*args, **kwargs)
    else:
        return urllib2.build_opener(*args, **kwargs)  # @UndefinedVariable


def to_timestamp(date_time):
    if PYTHON_33:
        return date_time.replace(tzinfo=datetime.timezone.utc).timestamp()
    else:
        return calendar.timegm(date_time.utctimetuple())


def to_datetime(timestamp):
    if PYTHON_33:
        return datetime.datetime.fromtimestamp(
            timestamp, datetime.timezone.utc
        ).replace(tzinfo=None)
    else:
        return datetime.datetime.utcfromtimestamp(timestamp)


def utcfromtimestamp(timestamp):
    return to_datetime(timestamp)


def utc_now():
    if PYTHON_33:
        return datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
    else:
        return datetime.datetime.utcnow()


def urlparse(*args, **kwargs):
    return _urlparse.urlparse(*args, **kwargs)


def urlunparse(*args, **kwargs):
    return _urlparse.urlunparse(*args, **kwargs)


def parse_qs(*args, **kwargs):
    return _urlparse.parse_qs(*args, **kwargs)


def urlencode(*args, **kwargs):
    if PYTHON_3:
        return urllib.parse.urlencode(*args, **kwargs)
    else:
        return urllib.urlencode(*args, **kwargs)  # @UndefinedVariable


def quote(*args, **kwargs):
    if PYTHON_3:
        return urllib.parse.quote(*args, **kwargs)
    else:
        return urllib.quote(*args, **kwargs)  # @UndefinedVariable


def quote_plus(*args, **kwargs):
    if PYTHON_3:
        return urllib.parse.quote_plus(*args, **kwargs)
    else:
        return urllib.quote_plus(*args, **kwargs)  # @UndefinedVariable


def unquote(*args, **kwargs):
    if PYTHON_3:
        return urllib.parse.unquote(*args, **kwargs)
    else:
        return urllib.unquote(*args, **kwargs)  # @UndefinedVariable


def unquote_plus(*args, **kwargs):
    if PYTHON_3:
        return urllib.parse.unquote_plus(*args, **kwargs)
    else:
        return urllib.unquote_plus(*args, **kwargs)  # @UndefinedVariable


def cmp_to_key(*args, **kwargs):
    if PYTHON_3:
        return dict(key=functools.cmp_to_key(*args, **kwargs))  # @UndefinedVariable
    else:
        return dict(cmp=args[0])


def tobytes(self, *args, **kwargs):
    if PYTHON_3:
        return self.tobytes(*args, **kwargs)
    else:
        return self.tostring(*args, **kwargs)


def tostring(self, *args, **kwargs):
    if PYTHON_3:
        return self.tobytes(*args, **kwargs)
    else:
        return self.tostring(*args, **kwargs)


def StringIO(*args, **kwargs):
    if PYTHON_3:
        return cStringIO.StringIO(*args, **kwargs)
    else:
        return _StringIO.StringIO(*args, **kwargs)


def BytesIO(*args, **kwargs):
    if PYTHON_3:
        return cStringIO.BytesIO(*args, **kwargs)
    else:
        return cStringIO.StringIO(*args, **kwargs)


class Orderable(tuple):
    """
    Simple tuple type wrapper that provides a simple
    first element ordering, that is compatible with
    both the Python 2 and Python 3+ infra-structures.
    """

    def __cmp__(self, value):
        return self[0].__cmp__(value[0])

    def __lt__(self, value):
        return self[0].__lt__(value[0])



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import unittest

import appier


class GraphTest(unittest.TestCase):
    def test__build_path(self):
        prev = dict(B="A", C="A", D="B", E="D", F="D", G="E")

        path = appier.Graph._build_path(prev, "A", "F")
        self.assertEqual(path, ["A", "B", "D", "F"])

    def test_create(self):
        graph = appier.Graph()
        self.assertEqual(len(graph.edges), 0)

    def test_create_with_argument(self):
        graph = appier.Graph([("A", "B"), ("B", "D", 20, True)])
        self.assertEqual(len(graph.edges), 3)

    def test_add_edges(self):
        graph = appier.Graph()
        edges = [
            ("A", "B"),
            ("B", "D", 20),
            ("B", "C", 10, True),
            ("D", "F"),
            ("F", "D"),
        ]
        graph.add_edges(edges)

        self.assertEqual(graph.edges["A"], [("B", 1)])
        self.assertEqual(graph.edges["B"], [("D", 20), ("C", 10)])
        self.assertEqual(graph.edges["D"], [("F", 1)])
        self.assertEqual(graph.edges["F"], [("D", 1)])

    def test_add_edges_handle_invalid(self):
        graph = appier.Graph()
        edges = [
            ("A", "B", "invalid", "invalid"),
            ("B", "D", 20),
            ("B", "C", 10, True),
            ("D", "F"),
            ("F", "D"),
            (),
            ("A"),
        ]
        graph.add_edges(edges)

        self.assertEqual(graph.edges["A"], [("B", 1)])
        self.assertEqual(graph.edges["B"], [("D", 20), ("C", 10)])
        self.assertEqual(graph.edges["D"], [("F", 1)])
        self.assertEqual(graph.edges["F"], [("D", 1)])

    def test_add_edge(self):
        graph = appier.Graph()

        graph.add_edge("A", "B")
        self.assertEqual(graph.edges["A"], [("B", 1)])

        graph.add_edge("B", "D", cost=20)
        graph.add_edge("B", "C", cost=10)
        self.assertEqual(graph.edges["B"], [("D", 20), ("C", 10)])

        graph.add_edge("D", "F", bidirectional=True)
        self.assertEqual(graph.edges["D"], [("F", 1)])
        self.assertEqual(graph.edges["F"], [("D", 1)])

    def test_disjktra_no_path(self):
        graph = appier.Graph([("A", "B"), ("B", "C"), ("F", "G")])

        path, cost = graph.dijkstra("C", "A")
        self.assertEqual(path, [])
        self.assertEqual(cost, appier.defines.INFINITY)

        path, cost = graph.dijkstra("C", "B")
        self.assertEqual(path, [])
        self.assertEqual(cost, appier.defines.INFINITY)

        path, cost = graph.dijkstra("A", "F")
        self.assertEqual(path, [])
        self.assertEqual(cost, appier.defines.INFINITY)

    def test_dijkstra_src_equal_dst(self):
        graph = appier.Graph()

        path, cost = graph.dijkstra("A", "A")
        self.assertEqual(path, ["A"])
        self.assertEqual(cost, 0)

    def test_dijkstra_simple(self):
        graph = appier.Graph([("A", "B"), ("B", "C")])

        path, cost = graph.dijkstra("A", "C")
        self.assertEqual(path, ["A", "B", "C"])
        self.assertEqual(cost, 2)

    def test_dijkstra_costs(self):
        graph = appier.Graph([("A", "B"), ("B", "C", 10), ("B", "D", 4), ("D", "C", 5)])

        path, cost = graph.dijkstra("A", "C")
        self.assertEqual(path, ["A", "B", "D", "C"])
        self.assertEqual(cost, 10)

    def test_dijkstra_loop(self):
        graph = appier.Graph([("A", "B"), ("B", "B"), ("B", "C")])

        path, cost = graph.dijkstra("A", "C")
        self.assertEqual(path, ["A", "B", "C"])
        self.assertEqual(cost, 2)

    def test_dijkstra_big(self):
        graph = appier.Graph(
            [
                ("A", "B", 2),
                ("A", "C", 6),
                ("B", "D", 5),
                ("C", "D", 8),
                ("D", "E", 10),
                ("D", "F", 15),
                ("E", "F", 6),
                ("E", "G", 2),
                ("F", "G", 6),
            ]
        )

        path, cost = graph.dijkstra("A", "A")
        self.assertEqual(path, ["A"])
        self.assertEqual(cost, 0)

        path, cost = graph.dijkstra("A", "B")
        self.assertEqual(path, ["A", "B"])
        self.assertEqual(cost, 2)

        path, cost = graph.dijkstra("A", "C")
        self.assertEqual(path, ["A", "C"])
        self.assertEqual(cost, 6)

        path, cost = graph.dijkstra("A", "D")
        self.assertEqual(path, ["A", "B", "D"])
        self.assertEqual(cost, 7)

        path, cost = graph.dijkstra("A", "E")
        self.assertEqual(path, ["A", "B", "D", "E"])
        self.assertEqual(cost, 17)

        path, cost = graph.dijkstra("A", "F")
        self.assertEqual(path, ["A", "B", "D", "F"])
        self.assertEqual(cost, 22)

        path, cost = graph.dijkstra("A", "G")
        self.assertEqual(path, ["A", "B", "D", "E", "G"])
        self.assertEqual(cost, 19)

        path, cost = graph.dijkstra("C", "G")
        self.assertEqual(path, ["C", "D", "E", "G"])
        self.assertEqual(cost, 20)

        path, cost = graph.dijkstra("C", "F")
        self.assertEqual(path, ["C", "D", "F"])
        self.assertEqual(cost, 23)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import unittest

import appier


class UtilTest(unittest.TestCase):
    def test_to_sort(self):
        result = appier.to_sort("name:descending")
        self.assertEqual(result, [("name", -1)])

        result = appier.to_sort("name:ascending")
        self.assertEqual(result, [("name", 1)])

        result = appier.to_sort("name:ascending,age:descending")
        self.assertEqual(result, [("name", 1), ("age", -1)])

    def test_is_mobile(self):
        result = appier.is_mobile(
            "Mozilla/5.0 (Linux; Android 4.0.4; Galaxy Nexus Build/IMM76B) AppleWebKit/535.19 (KHTML, like Gecko) Chrome/18.0.1025.133 Mobile Safari/535.19"
        )
        self.assertEqual(result, True)

        result = appier.is_mobile(
            "Mozilla/5.0 (iPhone; CPU iPhone OS 10_3 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) CriOS/56.0.2924.75 Mobile/14E5239e Safari/602.1"
        )
        self.assertEqual(result, True)

        result = appier.is_mobile(
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/600.7.12 (KHTML, like Gecko) Version/8.0.7 Safari/600.7.12"
        )
        self.assertEqual(result, False)

        result = appier.is_mobile(
            "Mozilla/5.0 (Linux; U; Android 4.1.1; en-gb; Build/KLP) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Safari/534.30"
        )
        self.assertEqual(result, False)

        result = appier.is_mobile(
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.75 Safari/537.36"
        )
        self.assertEqual(result, False)

        result = appier.is_mobile(
            "Mozilla/5.0 (iPad; CPU OS 9_3_2 like Mac OS X) AppleWebKit/601.1.46 (KHTML, like Gecko) Version/9.0 Mobile/13F69 Safari/601.1"
        )
        self.assertEqual(result, False)

        result = appier.is_mobile("")
        self.assertEqual(result, False)

        result = appier.is_mobile(None)
        self.assertEqual(result, False)

    def test_is_tablet(self):
        result = appier.is_tablet(
            "Mozilla/5.0 (iPad; CPU OS 9_3_2 like Mac OS X) AppleWebKit/601.1.46 (KHTML, like Gecko) Version/9.0 Mobile/13F69 Safari/601.1"
        )
        self.assertEqual(result, True)

        result = appier.is_tablet(
            "Mozilla/5.0 (iPad; CPU OS 6_1_3 like Mac OS X) AppleWebKit/536.26 (KHTML, like Gecko) Mobile/10B329"
        )
        self.assertEqual(result, True)

        result = appier.is_tablet(
            "Mozilla/5.0 (Linux; Android 4.0.4; Galaxy Nexus Build/IMM76B) AppleWebKit/535.19 (KHTML, like Gecko) Chrome/18.0.1025.133 Mobile Safari/535.19"
        )
        self.assertEqual(result, True)

        result = appier.is_tablet(
            "Mozilla/5.0 (iPhone; CPU iPhone OS 10_3 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) CriOS/56.0.2924.75 Mobile/14E5239e Safari/602.1"
        )
        self.assertEqual(result, True)

        result = appier.is_tablet(
            "Mozilla/5.0 (Linux; U; Android 4.1.1; en-gb; Build/KLP) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Safari/534.30"
        )
        self.assertEqual(result, True)

        result = appier.is_tablet(
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/600.7.12 (KHTML, like Gecko) Version/8.0.7 Safari/600.7.12"
        )
        self.assertEqual(result, False)

        result = appier.is_tablet(
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.75 Safari/537.36"
        )
        self.assertEqual(result, False)

        result = appier.is_tablet("")
        self.assertEqual(result, False)

        result = appier.is_tablet(None)
        self.assertEqual(result, False)

    def test_is_browser(self):
        result = appier.is_browser(
            "Mozilla/5.0 (iPad; CPU OS 9_3_2 like Mac OS X) AppleWebKit/601.1.46 (KHTML, like Gecko) Version/9.0 Mobile/13F69 Safari/601.1"
        )
        self.assertEqual(result, True)

        result = appier.is_browser(
            "Mozilla/5.0 (iPad; CPU OS 6_1_3 like Mac OS X) AppleWebKit/536.26 (KHTML, like Gecko) Mobile/10B329"
        )
        self.assertEqual(result, True)

        result = appier.is_browser(
            "Mozilla/5.0 (Linux; Android 4.0.4; Galaxy Nexus Build/IMM76B) AppleWebKit/535.19 (KHTML, like Gecko) Chrome/18.0.1025.133 Mobile Safari/535.19"
        )
        self.assertEqual(result, True)

        result = appier.is_browser(
            "Mozilla/5.0 (iPhone; CPU iPhone OS 10_3 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) CriOS/56.0.2924.75 Mobile/14E5239e Safari/602.1"
        )
        self.assertEqual(result, True)

        result = appier.is_browser(
            "Mozilla/5.0 (Linux; U; Android 4.1.1; en-gb; Build/KLP) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Safari/534.30"
        )
        self.assertEqual(result, True)

        result = appier.is_browser(
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/600.7.12 (KHTML, like Gecko) Version/8.0.7 Safari/600.7.12"
        )
        self.assertEqual(result, True)

        result = appier.is_browser(
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.75 Safari/537.36"
        )
        self.assertEqual(result, True)

        result = appier.is_browser(
            "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.135 Safari/537.36 Edge/12.10136"
        )
        self.assertEqual(result, True)

        result = appier.is_browser(
            "DuckDuckBot/1.0; (+http://duckduckgo.com/duckduckbot.html)"
        )
        self.assertEqual(result, False)

        result = appier.is_browser("netius/1.1.10")
        self.assertEqual(result, False)

        result = appier.is_browser("netius/1.1b")
        self.assertEqual(result, False)

        result = appier.is_browser("")
        self.assertEqual(result, False)

    def test_is_bot(self):
        result = appier.is_bot(
            "Mozilla/5.0 (iPad; CPU OS 9_3_2 like Mac OS X) AppleWebKit/601.1.46 (KHTML, like Gecko) Version/9.0 Mobile/13F69 Safari/601.1"
        )
        self.assertEqual(result, False)

        result = appier.is_bot(
            "Mozilla/5.0 (iPad; CPU OS 6_1_3 like Mac OS X) AppleWebKit/536.26 (KHTML, like Gecko) Mobile/10B329"
        )
        self.assertEqual(result, False)

        result = appier.is_bot(
            "Mozilla/5.0 (Linux; Android 4.0.4; Galaxy Nexus Build/IMM76B) AppleWebKit/535.19 (KHTML, like Gecko) Chrome/18.0.1025.133 Mobile Safari/535.19"
        )
        self.assertEqual(result, False)

        result = appier.is_bot(
            "Mozilla/5.0 (iPhone; CPU iPhone OS 10_3 like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko) CriOS/56.0.2924.75 Mobile/14E5239e Safari/602.1"
        )
        self.assertEqual(result, False)

        result = appier.is_bot(
            "Mozilla/5.0 (Linux; U; Android 4.1.1; en-gb; Build/KLP) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Safari/534.30"
        )
        self.assertEqual(result, False)

        result = appier.is_bot(
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/600.7.12 (KHTML, like Gecko) Version/8.0.7 Safari/600.7.12"
        )
        self.assertEqual(result, False)

        result = appier.is_bot(
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.75 Safari/537.36"
        )
        self.assertEqual(result, False)

        result = appier.is_bot(
            "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.135 Safari/537.36 Edge/12.10136"
        )
        self.assertEqual(result, False)

        result = appier.is_bot(
            "DuckDuckBot/1.0; (+http://duckduckgo.com/duckduckbot.html)"
        )
        self.assertEqual(result, True)

        result = appier.is_bot("netius/1.1.10")
        self.assertEqual(result, False)

        result = appier.is_bot("netius/1.1b")
        self.assertEqual(result, False)

        result = appier.is_bot("")
        self.assertEqual(result, False)

        result = appier.is_bot(None)
        self.assertEqual(result, False)

    def test_browser_info(self):
        result = appier.browser_info(
            "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.135 Safari/537.36 Edge/12.10136"
        )
        self.assertEqual(
            result,
            dict(
                name="Edge",
                version="12.10136",
                version_f=12.10136,
                version_i=12,
                interactive=True,
                bot=False,
                os="Windows",
            ),
        )

        result = appier.browser_info(
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.75 Safari/537.36"
        )
        self.assertEqual(
            result,
            dict(
                name="Chrome",
                version="62.0.3202.75",
                version_f=62.0,
                version_i=62,
                interactive=True,
                bot=False,
                os="Windows",
            ),
        )

        result = appier.browser_info(
            "Mozilla/5.0 (iPad; CPU OS 9_3_2 like Mac OS X) AppleWebKit/601.1.46 (KHTML, like Gecko) Version/9.0 Mobile/13F69 Safari/601.1"
        )
        self.assertEqual(
            result,
            dict(
                name="Safari",
                version="601.1",
                version_f=601.1,
                version_i=601,
                interactive=True,
                bot=False,
                os="Mac",
            ),
        )

        result = appier.browser_info(
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:56.0) Gecko/20100101 Firefox/56.0"
        )
        self.assertEqual(
            result,
            dict(
                name="Firefox",
                version="56.0",
                version_f=56.0,
                version_i=56,
                interactive=True,
                bot=False,
                os="Windows",
            ),
        )

        result = appier.browser_info(
            "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.0; Trident/4.0)"
        )
        self.assertEqual(
            result,
            dict(
                name="Explorer",
                version="8.0",
                version_f=8.0,
                version_i=8,
                interactive=True,
                bot=False,
                os="Windows",
            ),
        )

        result = appier.browser_info(
            "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)"
        )
        self.assertEqual(
            result,
            dict(
                name="Googlebot",
                version="2.1",
                version_f=2.1,
                version_i=2,
                interactive=False,
                bot=True,
            ),
        )

        result = appier.browser_info(
            "Mozilla/5.0 (compatible; Bingbot/2.0; +http://www.bing.com/bingbot.htm)"
        )
        self.assertEqual(
            result,
            dict(
                name="Bingbot",
                version="2.0",
                version_f=2.0,
                version_i=2,
                interactive=False,
                bot=True,
            ),
        )

        result = appier.browser_info(
            "DuckDuckBot/1.0; (+http://duckduckgo.com/duckduckbot.html)"
        )
        self.assertEqual(
            result,
            dict(
                name="DuckDuckBot",
                version="1.0",
                version_f=1.0,
                version_i=1,
                interactive=False,
                bot=True,
            ),
        )

        result = appier.browser_info("netius/1.1.10")
        self.assertEqual(
            result,
            dict(
                name="netius",
                version="1.1.10",
                version_f=1.1,
                version_i=1,
                interactive=False,
                bot=False,
            ),
        )

        result = appier.browser_info("netius/1.1b")
        self.assertEqual(
            result,
            dict(
                name="netius",
                version="1.1b",
                version_f=0,
                version_i=0,
                interactive=False,
                bot=False,
            ),
        )

        result = appier.browser_info(
            "APIs-Google (+https://developers.google.com/webmasters/APIs-Google.html)"
        )
        self.assertEqual(result, None)

        result = appier.browser_info(None)
        self.assertEqual(result, None)

    def test_obfuscate(self):
        result = appier.obfuscate("hello world")
        self.assertEqual(result, "hel********")

        result = appier.obfuscate("hello world", display_l=6)
        self.assertEqual(result, "hello *****")

        result = appier.obfuscate("hello world", display_l=100)
        self.assertEqual(result, "hello world")

        result = appier.obfuscate("hello world", display_l=6, token="-")
        self.assertEqual(result, "hello -----")

        result = appier.obfuscate(appier.legacy.u("你好世界"), display_l=3)
        self.assertEqual(result, appier.legacy.u("你好世*"))

    def test_email_parts(self):
        name, email = appier.email_parts("João Magalhães <joamag@hive.pt>")
        self.assertEqual(type(name), str)
        self.assertEqual(type(email), str)
        self.assertEqual(name, "João Magalhães")
        self.assertEqual(email, "joamag@hive.pt")

        name, email = appier.email_parts(
            appier.legacy.u("João Magalhães <joamag@hive.pt>")
        )
        self.assertEqual(type(name), appier.legacy.UNICODE)
        self.assertEqual(type(email), appier.legacy.UNICODE)
        self.assertEqual(name, appier.legacy.u("João Magalhães"))
        self.assertEqual(email, appier.legacy.u("joamag@hive.pt"))

        name, email = appier.email_parts(appier.legacy.u("你好世界 <hello@hive.pt>"))
        self.assertEqual(type(name), appier.legacy.UNICODE)
        self.assertEqual(type(email), appier.legacy.UNICODE)
        self.assertEqual(name, appier.legacy.u("你好世界"))
        self.assertEqual(email, appier.legacy.u("hello@hive.pt"))

        name, email = appier.email_parts(appier.legacy.u(" joamag@hive.pt "))
        self.assertEqual(type(name), appier.legacy.UNICODE)
        self.assertEqual(type(email), appier.legacy.UNICODE)
        self.assertEqual(name, appier.legacy.u("joamag@hive.pt"))
        self.assertEqual(email, appier.legacy.u("joamag@hive.pt"))

    def test_email_mime(self):
        result = appier.email_mime("João Magalhães <joamag@hive.pt>")
        self.assertEqual(type(result), str)
        self.assertEqual(
            result, "=?utf-8?q?Jo=C3=A3o_Magalh=C3=A3es?= <joamag@hive.pt>"
        )

        result = appier.email_mime(appier.legacy.u("João Magalhães <joamag@hive.pt>"))
        self.assertEqual(type(result), appier.legacy.UNICODE)
        self.assertEqual(
            result,
            appier.legacy.u("=?utf-8?q?Jo=C3=A3o_Magalh=C3=A3es?= <joamag@hive.pt>"),
        )

        result = appier.email_mime(appier.legacy.u(" joamag@hive.pt "))
        self.assertEqual(type(result), appier.legacy.UNICODE)
        self.assertEqual(result, appier.legacy.u("joamag@hive.pt <joamag@hive.pt>"))

        result = appier.email_mime(
            [
                appier.legacy.u("João Magalhães <joamag@hive.pt>"),
                appier.legacy.u(" joamag@hive.pt "),
                None,
            ]
        )
        self.assertEqual(type(result), list)
        self.assertEqual(
            result,
            [
                appier.legacy.u(
                    "=?utf-8?q?Jo=C3=A3o_Magalh=C3=A3es?= <joamag@hive.pt>"
                ),
                appier.legacy.u("joamag@hive.pt <joamag@hive.pt>"),
            ],
        )

    def test_email_name(self):
        result = appier.email_name("João Magalhães <joamag@hive.pt>")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "João Magalhães")

        result = appier.email_name(appier.legacy.u("João Magalhães <joamag@hive.pt>"))
        self.assertEqual(type(result), appier.legacy.UNICODE)
        self.assertEqual(result, appier.legacy.u("João Magalhães"))

        result = appier.email_name(appier.legacy.u(" joamag@hive.pt "))
        self.assertEqual(type(result), appier.legacy.UNICODE)
        self.assertEqual(result, appier.legacy.u("joamag@hive.pt"))

        result = appier.email_name(appier.legacy.u("joamag"))
        self.assertEqual(result, None)

        result = appier.email_name(appier.legacy.u(""))
        self.assertEqual(result, None)

        result = appier.email_name(
            [appier.legacy.u("joamag@hive.pt"), appier.legacy.u("joamag@hive.pt"), None]
        )
        self.assertEqual(type(result), list)
        self.assertEqual(
            result,
            [appier.legacy.u("joamag@hive.pt"), appier.legacy.u("joamag@hive.pt")],
        )

        result = appier.email_name(
            [appier.legacy.u("joamag"), appier.legacy.u("joamag@hive.pt"), None]
        )
        self.assertEqual(type(result), list)
        self.assertEqual(result, [appier.legacy.u("joamag@hive.pt")])

        result = appier.email_name([appier.legacy.u(""), None])
        self.assertEqual(type(result), list)
        self.assertEqual(result, [])

        result = appier.email_name([appier.legacy.u("")])
        self.assertEqual(type(result), list)
        self.assertEqual(result, [])

    def test_email_base(self):
        result = appier.email_base("João Magalhães <joamag@hive.pt>")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "joamag@hive.pt")

        result = appier.email_base(appier.legacy.u("João Magalhães <joamag@hive.pt>"))
        self.assertEqual(type(result), appier.legacy.UNICODE)
        self.assertEqual(result, appier.legacy.u("joamag@hive.pt"))

        result = appier.email_base(appier.legacy.u(" joamag@hive.pt "))
        self.assertEqual(type(result), appier.legacy.UNICODE)
        self.assertEqual(result, appier.legacy.u("joamag@hive.pt"))

        result = appier.email_base(appier.legacy.u("joamag"))
        self.assertEqual(result, None)

        result = appier.email_base(appier.legacy.u(""))
        self.assertEqual(result, None)

        result = appier.email_base(
            [appier.legacy.u("joamag@hive.pt"), appier.legacy.u("joamag@hive.pt"), None]
        )
        self.assertEqual(type(result), list)
        self.assertEqual(
            result,
            [appier.legacy.u("joamag@hive.pt"), appier.legacy.u("joamag@hive.pt")],
        )

        result = appier.email_base(
            [appier.legacy.u("joamag"), appier.legacy.u("joamag@hive.pt"), None]
        )
        self.assertEqual(type(result), list)
        self.assertEqual(result, [appier.legacy.u("joamag@hive.pt")])

        result = appier.email_base([appier.legacy.u(""), None])
        self.assertEqual(type(result), list)
        self.assertEqual(result, [])

        result = appier.email_base(
            [
                appier.legacy.u(""),
            ]
        )
        self.assertEqual(type(result), list)
        self.assertEqual(result, [])

    def test_date_to_timestamp(self):
        result = appier.date_to_timestamp("29/06/1984")
        self.assertEqual(type(result), int)
        self.assertEqual(int(result), 457315200)

        result = appier.date_to_timestamp("29/06/0000")
        self.assertEqual(result, None)

        result = appier.date_to_timestamp("1984-06-29", format="%Y-%m-%d")
        self.assertEqual(result, 457315200)

        result = appier.date_to_timestamp("1984-13-29", format="%Y-%m-%d")
        self.assertEqual(result, None)

    def test_gather_errors(self):
        def raiser():
            raise appier.OperationalError(message="hello")

        struct = appier.lazy_dict(
            first=appier.lazy(lambda: raiser()),
            second=appier.lazy(lambda: 2),
        )

        errors = appier.gather_errors(struct)
        self.assertEqual(errors, dict(first=["hello"]))

        struct.__getitem__("first", force=True)._value = 1

        errors = appier.gather_errors(struct)
        self.assertEqual(errors, dict(first=["hello"]))

        struct.__getitem__("first", force=True)._value = 1

        errors = appier.gather_errors(struct, resolve=False)
        self.assertEqual(errors, dict())

    def test_camel_to_underscore(self):
        result = appier.camel_to_underscore(None)
        self.assertEqual(result, None)

        result = appier.camel_to_underscore("")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "")

        result = appier.camel_to_underscore("HelloWorld")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "hello_world")

        result = appier.camel_to_underscore("HELLOWorld")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "hello_world")

        result = appier.camel_to_underscore("HELLOWorldHELLOWorld")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "hello_world_hello_world")

    def test_camel_to_readable(self):
        result = appier.camel_to_readable(None)
        self.assertEqual(result, None)

        result = appier.camel_to_readable("")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "")

        result = appier.camel_to_readable("HelloWorld")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello World")

        result = appier.camel_to_readable("HelloWorld", lower=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello world")

        result = appier.camel_to_readable("HelloWorld", lower=True, capitalize=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello World")

        result = appier.camel_to_readable("HELLOWorld")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "HELLO World")

        result = appier.camel_to_readable("HELLOWorld", lower=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello world")

        result = appier.camel_to_readable("HELLOWorld", lower=True, capitalize=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello World")

        result = appier.camel_to_readable("HELLOWorldHELLOWorld")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "HELLO World HELLO World")

        result = appier.camel_to_readable("HELLOWorldHELLOWorld", lower=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello world hello world")

        result = appier.camel_to_readable(
            "HELLOWorldHELLOWorld", lower=True, capitalize=True
        )
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello World Hello World")

    def test_underscore_to_camel(self):
        result = appier.underscore_to_camel(None)
        self.assertEqual(result, None)

        result = appier.underscore_to_camel("")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "")

        result = appier.underscore_to_camel("hello_world")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "HelloWorld")

        result = appier.underscore_to_camel("hello_world", lower=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "helloWorld")

        result = appier.underscore_to_camel("hello_world_hello_world")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "HelloWorldHelloWorld")

        result = appier.underscore_to_camel("hello_world_hello_world", lower=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "helloWorldHelloWorld")

        result = appier.underscore_to_camel("hello_world_")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "HelloWorld")

        result = appier.underscore_to_camel("hello_world_", lower=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "helloWorld")

        result = appier.underscore_to_camel("__hello_world__")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "HelloWorld")

        result = appier.underscore_to_camel("__hello_world__", lower=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "helloWorld")

        result = appier.underscore_to_camel("__hello___world__")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "HelloWorld")

        result = appier.underscore_to_camel("__hello___world__", lower=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "helloWorld")

        result = appier.underscore_to_camel("__hello___WORLD__")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "HelloWORLD")

        result = appier.underscore_to_camel("__hello___WORLD__", lower=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "helloWORLD")

        result = appier.underscore_to_camel("HelloWorld")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "HelloWorld")

        result = appier.underscore_to_camel("HelloWorld", lower=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "helloWorld")

    def test_underscore_to_readable(self):
        result = appier.underscore_to_readable(None)
        self.assertEqual(result, None)

        result = appier.underscore_to_readable("")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "")

        result = appier.underscore_to_readable("hello_world")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello world")

        result = appier.underscore_to_readable("hello_world", capitalize=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello World")

        result = appier.underscore_to_readable("hello_world_hello_world")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello world hello world")

        result = appier.underscore_to_readable(
            "hello_world_hello_world", capitalize=True
        )
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello World Hello World")

        result = appier.underscore_to_readable("hello_world_")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello world")

        result = appier.underscore_to_readable("hello_world_", capitalize=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello World")

        result = appier.underscore_to_readable("__hello_world__")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello world")

        result = appier.underscore_to_readable("__hello_world__", capitalize=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello World")

        result = appier.underscore_to_readable("__hello___world__")
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello world")

        result = appier.underscore_to_readable("__hello___world__", capitalize=True)
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello World")

        result = appier.underscore_to_readable(
            "__hello___world__", capitalize=True, separator="-"
        )
        self.assertEqual(type(result), str)
        self.assertEqual(result, "Hello-World")

    def test_escape(self):
        result = appier.escape("foo,bar", ",", escape="$")
        self.assertEqual(result, "foo$,bar")

        result = appier.escape("foo$,bar", ",", escape="$")
        self.assertEqual(result, "foo$$$,bar")

        result = appier.escape("foo$,bar:", (",", ":"), escape="$")
        self.assertEqual(result, "foo$$$,bar$:")

    def test_unescape(self):
        result = appier.unescape("foo$,bar", escape="$")
        self.assertEqual(result, "foo,bar")

        result = appier.unescape("foo$$,bar", escape="$")
        self.assertEqual(result, "foo$,bar")

        result = appier.unescape("foo$$,bar$:", escape="$")
        self.assertEqual(result, "foo$,bar:")

        result = appier.unescape("$$foo$,bar$$$$", escape="$")
        self.assertEqual(result, "$foo,bar$$")

    def test_count_unescape(self):
        result = appier.count_unescape("foo:bar", ":")
        self.assertEqual(result, 1)

        result = appier.count_unescape("foo:bar:hello:world", ":")
        self.assertEqual(result, 3)

        result = appier.count_unescape("foo,bar,hello,world", ":")
        self.assertEqual(result, 0)

        result = appier.count_unescape("foo:bar\\:hello:world", ":")
        self.assertEqual(result, 2)

        result = appier.count_unescape("foo:bar\\:hello\\:world", ":")
        self.assertEqual(result, 1)

        result = appier.count_unescape("foo:bar\\:hello\\\\:world", ":")
        self.assertEqual(result, 2)

        result = appier.count_unescape("foo\\:bar\\:hello\\:world", ":")
        self.assertEqual(result, 0)

    def test_split_unescape(self):
        result = appier.split_unescape("foo bar")
        self.assertEqual(result, ["foo", "bar"])

        result = appier.split_unescape("foo bar hello world", max=2)
        self.assertEqual(result, ["foo", "bar", "hello world"])

        result = appier.split_unescape("foo,bar", ",")
        self.assertEqual(result, ["foo", "bar"])

        result = appier.split_unescape("foo$,bar", ",", escape="$")
        self.assertEqual(result, ["foo,bar"])

        result = appier.split_unescape("foo$$,bar", ",", escape="$", unescape=True)
        self.assertEqual(result, ["foo$", "bar"])

        result = appier.split_unescape("foo$$,bar", ",", escape="$", unescape=False)
        self.assertEqual(result, ["foo$$", "bar"])

        result = appier.split_unescape("foo$", ",", escape="$", unescape=True)
        self.assertEqual(result, ["foo$"])

        result = appier.split_unescape("foo\\\\\\:bar", ":", unescape=True)
        self.assertEqual(result, ["foo\\:bar"])

        result = appier.split_unescape("foo\\\\:bar", ":", unescape=True)
        self.assertEqual(result, ["foo\\", "bar"])

    def test_is_content_type(self):
        result = appier.is_content_type("text/plain", "text/plain")
        self.assertEqual(result, True)

        result = appier.is_content_type("text/plain", ("text/plain",))
        self.assertEqual(result, True)

        result = appier.is_content_type("text/plain", "text/html")
        self.assertEqual(result, False)

        result = appier.is_content_type("text/plain", ("text/html",))
        self.assertEqual(result, False)

        result = appier.is_content_type("text/plain", ("text/plain", "text/html"))
        self.assertEqual(result, True)

        result = appier.is_content_type("text/*", "text/plain")
        self.assertEqual(result, True)

        result = appier.is_content_type("text/*", "text/json")
        self.assertEqual(result, True)

    def test_parse_content_type(self):
        result = appier.parse_content_type("text/plain")
        self.assertEqual(type(result), tuple)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], ["text/plain"])
        self.assertEqual(result[1], dict())

        result = appier.parse_content_type("text/plain+json")
        self.assertEqual(type(result), tuple)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], ["text/plain", "text/json"])
        self.assertEqual(result[1], dict())

        result = appier.parse_content_type("text/plain+json; charset=utf-8")
        self.assertEqual(type(result), tuple)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], ["text/plain", "text/json"])
        self.assertEqual(result[1], dict(charset="utf-8"))

        result = appier.parse_content_type("text/plain+json   ; charset=utf-8")
        self.assertEqual(type(result), tuple)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], ["text/plain", "text/json"])
        self.assertEqual(result[1], dict(charset="utf-8"))

        result = appier.parse_content_type(
            "text/plain+json; charset=utf-8; boundary=hello;"
        )
        self.assertEqual(type(result), tuple)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], ["text/plain", "text/json"])
        self.assertEqual(result[1], dict(charset="utf-8", boundary="hello"))

        result = appier.parse_content_type("")
        self.assertEqual(type(result), tuple)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], [])
        self.assertEqual(result[1], dict())

        result = appier.parse_content_type("text")
        self.assertEqual(type(result), tuple)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], [])
        self.assertEqual(result[1], dict())

        result = appier.parse_content_type("text/plain+json; charset")
        self.assertEqual(type(result), tuple)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], ["text/plain", "text/json"])
        self.assertEqual(result[1], dict())

    def test_check_login(self):
        request = appier.Request("GET", "/", session_c=appier.MemorySession)

        request.session["tokens"] = ["*"]
        result = appier.check_login(None, token="admin", request=request)
        self.assertEqual(result, True)
        self.assertEqual(request.session["tokens"], {"*": True})

        request.session["tokens"] = []
        result = appier.check_login(None, token="admin", request=request)
        self.assertEqual(result, False)
        self.assertEqual(request.session["tokens"], {})

        request.session["tokens"] = ["admin"]
        result = appier.check_login(None, token="admin", request=request)
        self.assertEqual(result, True)
        self.assertEqual(request.session["tokens"], {"admin": True})

        request.session["tokens"] = ["admin.read"]
        result = appier.check_login(None, token="admin", request=request)
        self.assertEqual(result, False)
        result = appier.check_login(None, token="admin.read", request=request)
        self.assertEqual(result, True)
        self.assertEqual(request.session["tokens"], {"admin": {"read": True}})

        request.session["tokens"] = ["admin.*"]
        result = appier.check_login(None, token="admin.read", request=request)
        self.assertEqual(result, True)
        self.assertEqual(request.session["tokens"], {"admin": {"*": True}})

        request.session["tokens"] = ["admin", "admin.write"]
        result = appier.check_login(None, token="admin.read", request=request)
        self.assertEqual(result, False)
        self.assertEqual(
            request.session["tokens"], {"admin": {"_": True, "write": True}}
        )

        request.session["tokens"] = ["admin.write", "admin.*"]
        result = appier.check_login(None, token="admin.read", request=request)
        self.assertEqual(result, True)
        self.assertEqual(
            request.session["tokens"], {"admin": {"write": True, "*": True}}
        )

        del request.session["tokens"]
        result = appier.check_login(None, token="admin.read", request=request)
        self.assertEqual(result, False)
        self.assertEqual("tokens" in request.session, False)

    def test_check_tokens(self):
        result = appier.check_tokens(None, ("admin", "user"), tokens_m={"*": True})
        self.assertEqual(result, True)

        result = appier.check_tokens(None, ("admin", "user"), tokens_m={})
        self.assertEqual(result, False)

        result = appier.check_tokens(None, ("admin", "user"), tokens_m={"admin": True})
        self.assertEqual(result, False)

    def test_check_token(self):
        result = appier.check_token(None, "admin", tokens_m={"*": True})
        self.assertEqual(result, True)

        result = appier.check_token(None, "admin", tokens_m={})
        self.assertEqual(result, False)

        result = appier.check_token(None, "admin", tokens_m={"admin": True})
        self.assertEqual(result, True)

        result = appier.check_token(
            None, "admin.read", tokens_m={"admin": {"read": True}}
        )
        self.assertEqual(result, True)

        result = appier.check_token(None, "admin", tokens_m={"admin": {"read": True}})
        self.assertEqual(result, False)

        result = appier.check_token(None, "admin.read", tokens_m={"admin": {"*": True}})
        self.assertEqual(result, True)

        result = appier.check_token(None, None, tokens_m={})
        self.assertEqual(result, True)

    def test_to_tokens_m(self):
        result = appier.to_tokens_m(["admin"])
        self.assertEqual(result, {"admin": True})

        result = appier.to_tokens_m(["admin", "admin.read"])
        self.assertEqual(result, {"admin": {"_": True, "read": True}})

        result = appier.to_tokens_m(["admin.read", "admin"])
        self.assertEqual(result, {"admin": {"_": True, "read": True}})

        result = appier.to_tokens_m(["admin", "admin.*"])
        self.assertEqual(result, {"admin": {"_": True, "*": True}})

    def test_dict_merge(self):
        first = dict(a="hello", b="world")
        second = dict(a="hello_new", b="world_new", c="other")

        result = appier.dict_merge(first, second)
        self.assertEqual(id(result) in (id(first), (id(second))), False)
        self.assertEqual(result["a"], "hello_new")
        self.assertEqual(result["b"], "world_new")
        self.assertEqual(result["c"], "other")

        result = appier.dict_merge(first, second, override=False)
        self.assertEqual(id(result) in (id(first), (id(second))), False)
        self.assertEqual(result["a"], "hello")
        self.assertEqual(result["b"], "world")
        self.assertEqual(result["c"], "other")

        first = dict(a=dict(a="hello", b="world", d="other", m=dict(a="hello")))
        second = dict(
            a=dict(a="hello_new", b="world_new", c="other", m=dict(b="world"))
        )

        result = appier.dict_merge(first, second)
        self.assertEqual(id(result) in (id(first), (id(second))), False)
        self.assertEqual(
            result["a"],
            dict(a="hello_new", b="world_new", c="other", m=dict(b="world")),
        )

        result = appier.dict_merge(first, second, recursive=True)
        self.assertEqual(id(result) in (id(first), (id(second))), False)
        self.assertEqual(
            result["a"],
            dict(
                a="hello_new",
                b="world_new",
                c="other",
                d="other",
                m=dict(a="hello", b="world"),
            ),
        )

        result = appier.dict_merge(first, second, override=False, recursive=True)
        self.assertEqual(id(result) in (id(first), (id(second))), False)
        self.assertEqual(
            result["a"],
            dict(
                a="hello", b="world", c="other", d="other", m=dict(a="hello", b="world")
            ),
        )

        first = {
            "info": {
                "personal": {
                    "general": {
                        "kind": "human",
                    }
                }
            }
        }
        second = {
            "info": {
                "personal": {"general": {"kind": "cat", "tail": "long", "meaw": 12}},
                "profile": "base",
            }
        }

        result = appier.dict_merge(first, second, override=False, recursive=True)
        self.assertEqual(id(result) in (id(first), (id(second))), False)
        self.assertEqual(
            result,
            {
                "info": {
                    "personal": {
                        "general": {"kind": "human", "tail": "long", "meaw": 12}
                    },
                    "profile": "base",
                }
            },
        )

        result = appier.dict_merge(
            first["info"], second["info"], override=False, recursive=True
        )
        self.assertEqual(id(result) in (id(first), (id(second))), False)
        self.assertEqual(
            result,
            {
                "personal": {"general": {"kind": "human", "tail": "long", "meaw": 12}},
                "profile": "base",
            },
        )

    def test_verify(self):
        result = appier.verify(1 == 1)
        self.assertEqual(result, None)

        result = appier.verify("hello" == "hello")
        self.assertEqual(result, None)

        self.assertRaises(appier.AssertionError, lambda: appier.verify(1 == 2))

        self.assertRaises(
            appier.OperationalError,
            lambda: appier.verify(1 == 2, exception=appier.OperationalError),
        )

    def test_verify_equal(self):
        result = appier.verify_equal(1, 1)
        self.assertEqual(result, None)

        result = appier.verify_equal("hello", "hello")
        self.assertEqual(result, None)

        self.assertRaises(appier.AssertionError, lambda: appier.verify_equal(1, 2))

        self.assertRaises(
            appier.OperationalError,
            lambda: appier.verify_equal(1, 2, exception=appier.OperationalError),
        )

    def test_verify_not_equal(self):
        result = appier.verify_not_equal(1, 2)
        self.assertEqual(result, None)

        result = appier.verify_not_equal("hello", "world")
        self.assertEqual(result, None)

        self.assertRaises(appier.AssertionError, lambda: appier.verify_not_equal(1, 1))

        self.assertRaises(
            appier.OperationalError,
            lambda: appier.verify_not_equal(1, 1, exception=appier.OperationalError),
        )

    def test_verify_type(self):
        result = appier.verify_type("hello", str)
        self.assertEqual(result, None)

        result = appier.verify_type(1, int)
        self.assertEqual(result, None)

        result = appier.verify_type(None, int)
        self.assertEqual(result, None)

        self.assertRaises(appier.AssertionError, lambda: appier.verify_type(1, str))

        self.assertRaises(
            appier.OperationalError,
            lambda: appier.verify_type(1, str, exception=appier.OperationalError),
        )

        self.assertRaises(
            appier.AssertionError, lambda: appier.verify_type(None, str, null=False)
        )

        self.assertRaises(
            appier.OperationalError,
            lambda: appier.verify_type(
                None, str, null=False, exception=appier.OperationalError
            ),
        )

    def test_verify_many(self):
        result = appier.verify_many((1 == 1, 2 == 2, 3 == 3))
        self.assertEqual(result, None)

        result = appier.verify_many(("hello" == "hello",))
        self.assertEqual(result, None)

        self.assertRaises(appier.AssertionError, lambda: appier.verify_many((1 == 2,)))

        self.assertRaises(
            appier.AssertionError, lambda: appier.verify_many((1 == 1, 1 == 2))
        )

        self.assertRaises(
            appier.OperationalError,
            lambda: appier.verify_many(
                (1 == 1, 1 == 2), exception=appier.OperationalError
            ),
        )


class FileTupleTest(unittest.TestCase):
    def test_basic(self):
        file = appier.FileTuple.from_data(
            b"hello world", name="hello", mime="text/plain"
        )

        self.assertEqual(file.read(), b"hello world")
        self.assertEqual(file.name, "hello")
        self.assertEqual(file.mime, "text/plain")
        self.assertEqual(file.data, b"hello world")

    def test_interface(self):
        file = appier.FileTuple.from_data(
            b"hello world", name="hello", mime="text/plain"
        )

        self.assertEqual(file.read(), b"hello world")
        self.assertEqual(file.tell(), 11)

        file.seek(0)

        self.assertEqual(file.tell(), 0)

        self.assertEqual(file.read(5), b"hello")
        self.assertEqual(file.tell(), 5)


class BaseThreadTest(unittest.TestCase):
    def test_basic(self):
        thread = appier.BaseThread(args=(), daemon=True, name="Test")

        self.assertEqual(thread.name, "Test")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import time
import heapq
import calendar
import datetime
import logging
import threading
import traceback

from . import config
from . import legacy

LOOP_TIMEOUT = 60.0
""" The time value to be used to sleep the main sequence
loop between ticks, this value should not be too small
to spend many resources or to high to create a long set
of time between external interactions """


class Scheduler(threading.Thread):
    """
    Scheduler class that handles timeout based async tasks
    within the context of an Appier application.

    The architecture of the logic for the class should be
    modular in the sense that new task may be added to
    it through a queue or other external system. For that
    a proper preemption mechanism should exist allowing
    the scheduler to be stopped and started again.
    """

    def __init__(self, owner, timeout=LOOP_TIMEOUT, daemon=True):
        threading.Thread.__init__(self, name="Scheduler")
        self.owner = owner
        self.timeout = config.conf("SCHEDULER_TIMEOUT", timeout, cast=float)
        self.daemon = config.conf("SCHEDULER_DAEMON", daemon, cast=bool)
        self._condition = threading.Condition()

    def run(self):
        self.running = True
        self.load()
        while self.running:
            try:
                self.tick()
            except Exception as exception:
                self.logger.critical("Unhandled scheduler exception raised")
                self.logger.error(exception)
                lines = traceback.format_exc().splitlines()
                for line in lines:
                    self.logger.warning(line)
            self._condition.acquire()
            self._condition.wait(self.timeout)
            self._condition.release()

    def stop(self, awake=True):
        self.running = False
        if awake:
            self.awake()

    def tick(self):
        pass

    def load(self):
        pass

    def awake(self):
        self._condition.acquire()
        self._condition.notify()
        self._condition.release()

    @property
    def logger(self):
        if self.owner:
            return self.owner.logger
        else:
            return logging.getLogger()


class CronScheduler(Scheduler):
    """
    Specialized version of the scheduler that runs tasks
    based on a cron like configuration.

    The tasks are defined in a cron like format and are
    executed based on the current time.
    """

    def __init__(self, owner, timeout=LOOP_TIMEOUT, daemon=True):
        Scheduler.__init__(self, owner, timeout=timeout, daemon=daemon)
        self._tasks = []

    def tick(self, now_ts=None):
        current_ts = lambda: now_ts if now_ts else time.time()
        current_dt = lambda: (
            legacy.to_datetime(now_ts) if now_ts else legacy.utc_now()
        )

        timestamp = current_ts() + 5.0

        while True:
            if not self._tasks:
                break

            timestamp, task = self._tasks[0]
            if timestamp > current_ts():
                break

            heapq.heappop(self._tasks)

            if task.enabled:
                task.job()
                heapq.heappush(
                    self._tasks, (task.next_timestamp(now=current_dt()), task)
                )

        self.timeout = max(0, timestamp - current_ts())

    def schedule(self, job, cron, now=None):
        """
        Schedules the provided job function for execution according
        to the provided cron string.

        The optional now parameter may be used to provide the current
        time reference for the scheduling operation, meaning that the next
        timestamp will be calculated using this value as reference.

        :type job: Function
        :param job: The function to be executed as the job.
        :type cron: String/SchedulerDate
        :param cron: The cron like string defining the schedule.
        :type now: datetime
        :param now: Optional time reference for the job scheduling.
        :rtype: SchedulerTask
        :return: The task object that was created for the job.
        """

        task = SchedulerTask(job, cron)
        heapq.heappush(self._tasks, (task.next_timestamp(now=now), task))
        self.awake()
        return task

    def next_run(self):
        timestamp = self.next_timestamp()
        if not timestamp:
            return None
        return legacy.to_datetime(timestamp)

    def next_timestamp(self):
        if not self._tasks:
            return None
        return self._tasks[0][0]


class SchedulerTask(object):

    def __init__(self, job, cron):
        self.job = job
        self.date = SchedulerDate.from_cron(cron)
        self._enabled = True

    def enable(self):
        self._enabled = True

    def disable(self):
        self._enabled = False

    def next_run(self, now=None):
        return self.date.next_run(now=now)

    def next_timestamp(self, now=None):
        return self.date.next_timestamp(now=now)

    @property
    def enabled(self):
        return self._enabled


class SchedulerDate(object):

    def __init__(
        self, minutes="*", hours="*", days_of_month="*", months="*", days_of_week="*"
    ):
        self.minutes = self._parse_field(minutes, 0, 59)
        self.hours = self._parse_field(hours, 0, 23)
        self.days_of_month = self._parse_field(days_of_month, 1, 31)
        self.months = self._parse_field(months, 1, 12)
        self.days_of_week = self._parse_field(days_of_week, 0, 6)

    @classmethod
    def from_cron(cls, cron):
        if isinstance(cron, cls):
            return cron
        values = (value.strip().split(",") for value in cron.split(" "))
        return cls(*values)

    def next_timestamp(self, now=None):
        date = self.next_run(now=now)
        return legacy.to_timestamp(date)

    def next_run(self, now=None):
        """
        Calculate the next run time starting from the current time.
        This operation is done respecting Cron rules.

        :type now: datetime
        :param now: Optional date time to be used as the current time.
        :rtype: datetime
        :return: The next run time respecting Cron rules.
        """

        now = now or legacy.utc_now()
        now_day = datetime.datetime(now.year, now.month, now.day)
        now_hour = datetime.datetime(now.year, now.month, now.day, hour=now.hour)
        now_minute = datetime.datetime(
            now.year, now.month, now.day, hour=now.hour, minute=now.minute
        )

        year = now.year

        while True:
            for month in sorted(self.months):
                if month < now.month and year < now.year:
                    continue

                for day in sorted(self.days_of_month):
                    try:
                        date = datetime.datetime(year, month, day)
                    except ValueError:
                        continue
                    if self.days_of_week and not date.weekday() in self.days_of_week:
                        continue
                    if date < now_day:
                        continue

                    for hour in sorted(self.hours):
                        if date.replace(hour=hour) < now_hour:
                            continue

                        for minute in sorted(self.minutes):
                            _date = date.replace(
                                hour=hour, minute=minute, second=0, microsecond=0
                            )
                            if _date > now_minute:
                                return _date

            year += 1

    def _parse_field(self, field, min_value, max_value):
        if field in ("*", ["*"], ("*",)):
            return set(range(min_value, max_value + 1))
        elif isinstance(field, (list, tuple)):
            return set(int(v) for v in field)
        else:
            return set((int(field),))


class Cron(object):
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import copy

from . import common
from . import legacy
from . import exceptions


class ModelAsync(object):
    @classmethod
    async def get_a(cls, *args, **kwargs):
        (
            fields,
            eager,
            eager_l,
            map,
            rules,
            meta,
            build,
            fill,
            resolve_a,
            skip,
            limit,
            sort,
            raise_e,
        ) = cls._get_attrs(
            kwargs,
            (
                ("fields", None),
                ("eager", None),
                ("eager_l", None),
                ("map", False),
                ("rules", True),
                ("meta", False),
                ("build", True),
                ("fill", True),
                ("resolve_a", None),
                ("skip", 0),
                ("limit", 0),
                ("sort", None),
                ("raise_e", True),
            ),
        )

        # in case there's a sort field and the safe search mode is enabled
        # we must add sorting by the `_id` field so that the retrieval is
        # considered to be deterministic, otherwise some DB implementations
        # will not respect the same sorting sequence across different calls
        if sort and (skip or limit):
            if not isinstance(sort, list):
                sort = list(sort)
            sort.append(["_id", 1])

        if eager_l == None:
            eager_l = map
        if resolve_a == None:
            resolve_a = map
        if eager_l:
            eager = cls._eager_b(eager)
        fields = cls._sniff(fields, rules=rules)
        collection = cls._collection_a()
        model = await collection.find_one(
            kwargs, fields, skip=skip, limit=limit, sort=sort
        )
        if not model and raise_e:
            is_devel = common.is_devel()
            if is_devel:
                message = "%s not found for %s" % (cls.__name__, str(kwargs))
            else:
                message = "%s not found" % cls.__name__
            raise exceptions.NotFoundError(message=message)
        if not model and not raise_e:
            return model
        cls.types(model)
        if fill:
            cls.fill(model, safe=True)
        if build:
            cls.build(model, map=map, rules=rules, meta=meta)
        if eager:
            model = cls._eager(model, eager, map=map)
        if resolve_a:
            model = cls._resolve_all(model, resolve=False)
        return model if map else cls.old(model=model, safe=False)

    @classmethod
    async def find_a(cls, *args, **kwargs):
        (
            fields,
            eager,
            eager_l,
            map,
            rules,
            meta,
            build,
            fill,
            resolve_a,
            skip,
            limit,
            sort,
            raise_e,
        ) = cls._get_attrs(
            kwargs,
            (
                ("fields", None),
                ("eager", None),
                ("eager_l", False),
                ("map", False),
                ("rules", True),
                ("meta", False),
                ("build", True),
                ("fill", True),
                ("resolve_a", None),
                ("skip", 0),
                ("limit", 0),
                ("sort", None),
                ("raise_e", False),
            ),
        )

        # in case there's a sort field and the safe search mode is enabled
        # we must add sorting by the `_id` field so that the search is
        # considered to be deterministic, otherwise some DB implementations
        # will not respect the same sorting sequence across different calls
        if sort and (skip or limit):
            if not isinstance(sort, list):
                sort = list(sort)
            sort.append(["_id", 1])

        if resolve_a == None:
            resolve_a = map
        if eager_l:
            eager = cls._eager_b(eager)

        cls._find_s(kwargs)
        cls._find_d(kwargs)

        fields = cls._sniff(fields, rules=rules)
        collection = cls._collection_a()
        models = collection.find(kwargs, fields, skip=skip, limit=limit, sort=sort)
        if not models and raise_e:
            is_devel = common.is_devel()
            if is_devel:
                message = "%s not found for %s" % (cls.__name__, str(kwargs))
            else:
                message = "%s not found" % cls.__name__
            raise exceptions.NotFoundError(message=message)
        models = [cls.types(model) async for model in models]
        if fill:
            models = [cls.fill(model, safe=True) for model in models]
        if build:
            [cls.build(model, map=map, rules=rules, meta=meta) for model in models]
        if eager:
            models = cls._eager(models, eager, map=map)
        if resolve_a:
            models = [cls._resolve_all(model, resolve=False) for model in models]
        models = (
            models if map else [cls.old(model=model, safe=False) for model in models]
        )
        return models

    @classmethod
    async def _increment_a(cls, name):
        _name = cls._name() + ":" + name
        store = cls._collection_a(name="counters")
        value = await store.find_and_modify(
            {"_id": _name}, {"$inc": {"seq": 1}}, new=True, upsert=True
        )
        value = value or await store.find_one({"_id": _name})
        return value["seq"]

    async def save_a(
        self,
        validate=True,
        verify=True,
        is_new=None,
        increment_a=None,
        immutables_a=None,
        pre_validate=True,
        pre_save=True,
        pre_create=True,
        pre_update=True,
        post_validate=True,
        post_save=True,
        post_create=True,
        post_update=True,
        before_callbacks=[],
        after_callbacks=[],
    ):
        # ensures that the current instance is associated with
        # a concrete model, ready to be persisted in database
        if verify:
            self.assert_is_concrete()

        # checks if the instance to be saved is a new instance
        # or if this is an update operation and then determines
        # series of default values taking that into account
        if is_new == None:
            is_new = self.is_new()
        if increment_a == None:
            increment_a = is_new
        if immutables_a == None:
            immutables_a = not is_new

        # runs the validation process in the current model, this
        # should ensure that the model is ready to be saved in the
        # data source, without corruption of it, only run this process
        # in case the validate flag is correctly set
        validate and self._validate(
            pre_validate=pre_validate, post_validate=post_validate
        )

        # calls the complete set of event handlers for the current
        # save operation, this should trigger changes in the model
        if pre_save:
            self.pre_save()
        if pre_create and is_new:
            self.pre_create()
        if pre_update and not is_new:
            self.pre_update()

        # filters the values that are present in the current model
        # so that only the valid ones are stored in, invalid values
        # are going to be removed, note that if the operation is an
        # update operation and the "immutable rules" also apply, the
        # returned value is normalized meaning that for instance if
        # any relation is loaded the reference value is returned instead
        # of the loaded relation values (required for persistence)
        model = await self._filter_a(
            increment_a=increment_a, immutables_a=immutables_a, normalize=True
        )

        # in case the current model is not new must create a new
        # model instance and remove the main identifier from it
        if not is_new:
            _model = copy.copy(model)
            del _model["_id"]

        # calls the complete set of callbacks that should be called
        # before the concrete data store save operation
        for callback in before_callbacks:
            callback(self, model)

        # retrieves the reference to the store object to be used and
        # uses it to store the current model data
        store = self._get_store_a()
        if is_new:
            await store.insert(model)
            self.apply(model, safe_a=False)
        else:
            await store.update({"_id": model["_id"]}, {"$set": _model})

        # calls the complete set of callbacks that should be called
        # after the concrete data store save operation
        for callback in after_callbacks:
            callback(self, model)

        # calls the post save event handlers in order to be able to
        # execute appropriate post operations
        if post_save:
            self.post_save()
        if post_create and is_new:
            self.post_create()
        if post_update and not is_new:
            self.post_update()

        # returns the instance that has just been used for the save
        # operation, this may be used for chaining operations
        return self

    async def delete_a(
        self,
        verify=True,
        pre_delete=True,
        post_delete=True,
        before_callbacks=[],
        after_callbacks=[],
    ):
        # ensures that the current instance is associated with
        # a concrete model, ready to be persisted in database
        if verify:
            self.assert_is_concrete()

        # calls the complete set of event handlers for the current
        # delete operation, this should trigger changes in the model
        if pre_delete:
            self.pre_delete()

        # calls the complete set of callbacks that should be called
        # before the concrete data store delete operation
        for callback in before_callbacks:
            callback(self)

        # retrieves the reference to the store object to be able to
        # execute the removal command for the current model
        store = self._get_store_a()
        await store.remove({"_id": self._id})

        # calls the underlying delete handler that may be used to extend
        # the default delete functionality
        self._delete()

        # calls the complete set of callbacks that should be called
        # after the concrete data store delete operation
        for callback in after_callbacks:
            callback(self)

        # calls the complete set of event handlers for the current
        # delete operation, this should trigger changes in the model
        if post_delete:
            self.post_delete()

    async def reload_a(self, *args, **kwargs):
        is_new = self.is_new()
        if is_new:
            raise exceptions.OperationalError(
                message="Can't reload a new model entity", code=412
            )
        cls = self.__class__
        return await cls.get_a(_id=self._id, *args, **kwargs)

    async def _filter_a(
        self,
        increment_a=True,
        immutables_a=False,
        normalize=False,
        resolve=False,
        all=False,
        evaluator="json_v",
    ):
        # creates the model that will hold the "filtered" model
        # with all the items that conform with the class specification
        model = {}

        # retrieves the class associated with the current instance
        # to be able to retrieve the correct definition methods
        cls = self.__class__

        # retrieves the (schema) definition for the current model
        # to be "filtered" it's going to be used to retrieve the
        # various definitions for the model fields
        definition = cls.definition()

        # retrieves the complete list of fields that are meant to be
        # automatically incremented for every save operation
        increments = cls.increments()

        # gather the set of elements that are considered immutables and
        # that are not meant to be changed if the current operation to
        # apply the filter is not a new operation (update operation)
        immutables = cls.immutables()

        # iterates over all the increment fields and increments their
        # fields so that a new value is set on the model, note that if
        # the increment apply is unset the increment operation is ignored
        for name in increments:
            if not increment_a:
                continue
            if name in self.model:
                model[name] = cls._ensure_min(name, self.model[name])
            else:
                model[name] = await cls._increment_a(name)

        # iterates over all the model items to filter the ones
        # that are not valid for the current class context
        for name, value in legacy.eager(self.model.items()):
            if not name in definition:
                continue
            if immutables_a and name in immutables:
                continue
            value = self._evaluate(name, value, evaluator=evaluator)
            model[name] = value

        # in case the normalize flag is set must iterate over all
        # items to try to normalize the values by calling the reference
        # value this will returns the reference index value instead of
        # the normal value that would prevent normalization
        if normalize:
            for name, value in legacy.eager(self.model.items()):
                if not name in definition:
                    continue
                if not hasattr(value, "ref_v"):
                    continue
                model[name] = value.ref_v()

        # in case the resolution flag is set, it means that a recursive
        # approach must be performed for the resolution of values that
        # implement the map value (recursive resolution) method, this is
        # a complex (and possible computational expensive) process that
        # may imply access to the base data source
        if resolve:
            for name, value in legacy.eager(self.model.items()):
                if not name in definition:
                    continue
                model[name] = cls._resolve(name, value)

        # in case the all flag is set the extra fields (not present
        # in definition) must also be used to populate the resulting
        # (filtered) map so that it contains the complete set of values
        # present in the base map of the current instance
        if all:
            for name, value in legacy.eager(self.model.items()):
                if name in model:
                    continue
                model[name] = value

        # returns the model containing the "filtered" items resulting
        # from the validation of the items against the model class
        return model



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import json

from . import util
from . import legacy
from . import typesf
from . import config
from . import common
from . import exceptions

try:
    import pymongo
except ImportError:
    pymongo = None

try:
    import bson.json_util
except ImportError:
    bson = None

try:
    import motor.motor_asyncio
except ImportError:
    motor = None

URL = "mongodb://localhost"
""" The default URL to be used for the connection when
no other URL is provided (used most of the times) """

connection = None
""" The global connection object that should persist
the connection relation with the database service """

connection_a = None
""" The global connection reference for the async version
of the Mongo client """


class Mongo(object):
    def __init__(self, url=None):
        self.url = url
        self._connection = None
        self._db = None

    def get_connection(self, url=None, connect=False):
        if self._connection:
            return self._connection
        url_c = config.conf("MONGOHQ_URL", None)
        url_c = config.conf("MONGOLAB_URI", url_c)
        url_c = config.conf("MONGO_URL", url_c)
        url = url or self.url or url_c or URL
        if is_new():
            self._connection = _pymongo().MongoClient(url, connect=connect)
        else:
            self._connection = _pymongo().Connection(url)
        return self._connection

    def reset_connection(self):
        if not self._connection:
            return
        if is_new():
            self._connection.close()
        else:
            self._connection.disconnect()
        self._connection = None

    def get_db(self, name):
        if self._db:
            return self._db
        connection = self.get_connection()
        self._db = connection[name]
        return self._db


class MongoAsync(object):
    def __init__(self, url=None):
        self.url = url
        self._connection = None
        self._db = None

    def get_connection(self, url=None, connect=False):
        if self._connection:
            return self._connection
        url_c = config.conf("MONGOHQ_URL", None)
        url_c = config.conf("MONGOLAB_URI", url_c)
        url_c = config.conf("MONGO_URL", url_c)
        url = url or self.url or url_c or URL
        self._connection = _motor().AsyncIOMotorClient(url)
        return self._connection

    def reset_connection(self):
        if not self._connection:
            return
        self._connection.disconnect()
        self._connection = None

    def get_db(self, name):
        if self._db:
            return self._db
        connection = self.get_connection()
        self._db = connection[name]
        return self._db


class MongoEncoder(json.JSONEncoder):
    def default(self, obj, **kwargs):
        if not bson:
            return json.JSONEncoder.default(self, obj, **kwargs)
        if isinstance(obj, bson.objectid.ObjectId):
            return str(obj)
        if isinstance(obj, legacy.BYTES):
            return legacy.str(obj, encoding="utf-8")
        else:
            return json.JSONEncoder.default(self, obj, **kwargs)


def get_connection(url=URL, connect=False):
    global connection
    if connection:
        return connection
    url = config.conf("MONGOHQ_URL", url)
    url = config.conf("MONGOLAB_URI", url)
    url = config.conf("MONGO_URL", url)
    if is_new():
        connection = _pymongo().MongoClient(url, connect=connect)
    else:
        connection = _pymongo().Connection(url)
    return connection


def get_connection_a(url=URL, connect=False):
    global connection_a
    if connection_a:
        return connection_a
    url = config.conf("MONGOHQ_URL", url)
    url = config.conf("MONGOLAB_URI", url)
    url = config.conf("MONGO_URL", url)
    connection_a = _motor().AsyncIOMotorClient(url)
    return connection_a


def reset_connection():
    global connection
    if not connection:
        return
    if is_new():
        connection.close()
    else:
        connection.disconnect()
    connection = None


def reset_connection_a():
    global connection_a
    if not connection_a:
        return
    connection_a.close()
    connection_a = None


def get_db(name=None, get_connection=get_connection):
    url = config.conf("MONGOHQ_URL", None)
    url = config.conf("MONGOLAB_URI", url)
    url = config.conf("MONGO_URL", url)
    result = legacy.urlparse(url or "")
    name = result.path.strip("/") if result.path else None
    name = name or config.conf("MONGO_DB", name)
    name = name or common.base().get_name() or "master"
    connection = get_connection()
    db = connection[name]
    return db


def get_db_a(name=None, get_connection=get_connection):
    url = config.conf("MONGOHQ_URL", None)
    url = config.conf("MONGOLAB_URI", url)
    url = config.conf("MONGO_URL", url)
    result = legacy.urlparse(url or "")
    name = result.path.strip("/") if result.path else None
    name = name or config.conf("MONGO_DB", name)
    name = name or common.base().get_name() or "master"
    connection = get_connection_a()
    db = connection[name]
    return db


def drop_db(name=None, get_connection=get_connection):
    db = get_db(name=name)
    names = _list_names(db)
    for name in names:
        if name.startswith("system."):
            continue
        db.drop_collection(name)
    connection = get_connection()
    connection.drop_database(db.name)


def drop_db_a(name=None, get_connection=get_connection):
    db = get_db_a(name=name)
    names = _list_names(db)
    for name in names:
        if name.startswith("system."):
            continue
        db.drop_collection(name)
    connection = get_connection_a()
    connection.drop_database(db.name)


def object_id(value):
    return bson.ObjectId(value)


def dumps(*args):
    return json.dumps(default=serialize, *args)


def serialize(obj):
    if isinstance(obj, common.model().Model):
        return obj.model
    if isinstance(obj, typesf.AbstractType):
        return obj.json_v()
    return bson.json_util.default(obj)


def directions(all=False):
    return (
        (_pymongo().ASCENDING, _pymongo().DESCENDING, _pymongo().HASHED)
        if all
        else (_pymongo().ASCENDING, _pymongo().DESCENDING)
    )


def is_mongo(obj):
    if bson and isinstance(obj, bson.ObjectId):
        return True
    if bson and isinstance(obj, bson.DBRef):
        return True
    return False


def is_new(major=3, minor=0, patch=0):
    _major, _minor, _patch = _version_t()
    if _major > major:
        return True
    elif _major < major:
        return False
    if _minor > minor:
        return True
    elif _minor < minor:
        return False
    if _patch >= patch:
        return True
    else:
        return False


def _list_names(db, *args, **kwargs):
    if is_new(3, 7):
        return db.list_collection_names()
    else:
        return db.collection_names()


def _count(store, *args, **kwargs):
    if len(args) == 0:
        args = [{}]
    if is_new(3, 7):
        return store.count_documents(*args, **kwargs)
    return store.count(*args, **kwargs)


def _count_documents(store, *args, **kwargs):
    if len(args) == 0:
        args = [{}]
    if is_new(3, 7):
        return store.count_documents(*args, **kwargs)
    result = store.find(*args, **kwargs)
    return result.count()


def _store_find_and_modify(store, *args, **kwargs):
    if is_new():
        return store.find_one_and_update(*args, **kwargs)
    else:
        return store.find_and_modify(*args, **kwargs)


def _store_insert(store, *args, **kwargs):
    if is_new():
        return store.insert_one(*args, **kwargs)
    else:
        return store.insert(*args, **kwargs)


def _store_update(store, *args, **kwargs):
    if is_new():
        return store.update_one(*args, **kwargs)
    else:
        return store.update(*args, **kwargs)


def _store_remove(store, *args, **kwargs):
    if is_new():
        return store.delete_many(*args, **kwargs)
    else:
        return store.remove(*args, **kwargs)


def _store_ensure_index(store, *args, **kwargs):
    kwargs["background"] = kwargs.get("background", True)
    if is_new():
        return store.create_index(*args, **kwargs)
    else:
        return store.ensure_index(*args, **kwargs)


def _store_ensure_index_many(store, *args, **kwargs):
    directions_l = kwargs.pop("directions", None)
    if directions_l == "all":
        directions_l = directions(all=True)
    elif directions_l == None:
        directions_l = directions()
    for direction in directions_l:
        _args = list(args)
        _args[0] = [(_args[0], direction)]
        _store_ensure_index(store, *_args, **kwargs)


def _version_t():
    pymongo_l = _pymongo()
    if hasattr(pymongo_l, "_version_t"):
        return pymongo_l._version_t
    version_l = pymongo.version.split(".", 2)
    if len(version_l) == 2:
        version_l.append("0")
    major_s, minor_s, patch_s = version_l
    pymongo_l._version_t = (int(major_s), int(minor_s), int(patch_s))
    return pymongo_l._version_t


def _pymongo(verify=True):
    if verify:
        util.verify(
            not pymongo == None,
            message="PyMongo library not available",
            exception=exceptions.OperationalError,
        )
    return pymongo


def _motor(verify=True):
    if verify:
        util.verify(
            not motor == None,
            message="Motor library not available",
            exception=exceptions.OperationalError,
        )
    return motor.motor_asyncio



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Appier Framework
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Appier Framework.
#
# Hive Appier Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Appier Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Appier Framework. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import time
import pickle
import shutil

from . import util
from . import legacy
from . import common
from . import config
from . import redisdb
from . import component


class Cache(component.Component):
    def __init__(self, name="cache", owner=None, *args, **kwargs):
        component.Component.__init__(self, name=name, owner=owner, *args, **kwargs)
        load = kwargs.pop("load", True)
        if load:
            self.load(*args, **kwargs)

    def __len__(self):
        return self.length()

    def __getitem__(self, key):
        return self.get_item(key)

    def __setitem__(self, key, value):
        return self.set_item(key, value)

    def __delitem__(self, key):
        return self.delete_item(key)

    def __contains__(self, item):
        return self.contains(item)

    def __nonzero__(self):
        return True

    def __bool__(self):
        return True

    @classmethod
    def new(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def length(self):
        return 0

    def clear(self):
        pass

    def get(self, key, default=None):
        try:
            return self.get_item(key)
        except KeyError:
            return default

    def get_item(self, key):
        raise KeyError("not found")

    def set(self, *args, **kwargs):
        return self.set_item(*args, **kwargs)

    def set_item(self, key, value, expires=None, timeout=None):
        self.mark()

    def delete_item(self, key):
        self.mark()

    def contains(self, key):
        try:
            self.get_item(key)
        except KeyError:
            return False
        return True

    def build_value(self, value, expires, timeout):
        if timeout:
            expires = time.time() + timeout
        return (value, expires)


class MemoryCache(Cache):
    def __init__(self, name="memory", owner=None, *args, **kwargs):
        Cache.__init__(self, name=name, owner=owner, *args, **kwargs)

    def length(self):
        return self._data.__len__()

    def clear(self):
        self._data.clear()

    def get_item(self, key):
        value, expires = self._data.__getitem__(key)
        if not expires == None and expires < time.time():
            self.delete_item(key)
            return self.get_item(key)
        return value

    def set_item(self, key, value, expires=None, timeout=None):
        value = self.build_value(value, expires, timeout)
        return self._data.__setitem__(key, value)

    def delete_item(self, key):
        return self._data.__delitem__(key)

    def _load(self, *args, **kwargs):
        Cache._load(self, *args, **kwargs)
        self._data = dict()

    def _unload(self, *args, **kwargs):
        Cache._unload(self, *args, **kwargs)
        self._data = None


class FileCache(Cache):
    def __init__(self, name="file", owner=None, *args, **kwargs):
        Cache.__init__(self, name=name, owner=owner, *args, **kwargs)

    def length(self):
        if not os.path.exists(self.base_path):
            return 0
        return len(os.listdir(self.base_path))

    def clear(self):
        if not os.path.exists(self.base_path):
            return
        shutil.rmtree(self.base_path, ignore_errors=True)

    def get_item(self, key):
        file_path = os.path.join(self.base_path, key)
        if not os.path.exists(file_path):
            raise KeyError("not found")
        if not os.path.exists(file_path + ".expires"):
            raise KeyError("not found")
        expires = self._read_file(file_path + ".expires")
        expires = int(expires) if expires else None
        if not expires == None and expires < time.time():
            self.delete_item(key)
            return self.get_item(key)
        value = self._read_file(file_path)
        return value

    def set_item(self, key, value, expires=None, timeout=None):
        file_path = os.path.join(self.base_path, key)
        value, expires = self.build_value(value, expires, timeout)
        expires = str(int(expires)) if expires else None
        expires_s = legacy.bytes(expires, force=True) if expires else b""
        self._write_file(file_path, value)
        self._write_file(file_path + ".expires", expires_s)

    def delete_item(self, key):
        file_path = os.path.join(self.base_path, key)
        os.remove(file_path)
        os.remove(file_path + ".expires")

    def _load(self, *args, **kwargs):
        Cache._load(self, *args, **kwargs)
        self.base_path = kwargs.pop("base_path", None)
        self._ensure_path()

    def _read_file(self, file_path):
        file = open(file_path, "rb")
        try:
            data = file.read()
        finally:
            file.close()
        return data

    def _write_file(self, file_path, data):
        self._ensure_exists()
        file = open(file_path, "wb")
        try:
            file.write(data)
        finally:
            file.close()

    def _ensure_exists(self):
        if os.path.exists(self.base_path):
            return
        os.makedirs(self.base_path)

    def _ensure_path(self):
        if self.base_path:
            return
        app_path = common.base().get_base_path()
        util.verify(not app_path == None, message="No app path available")
        cache_path = os.path.join(app_path, "cache")
        cache_path = config.conf("CACHE_PATH", cache_path)
        cache_path = os.path.expanduser(cache_path)
        cache_path = os.path.abspath(cache_path)
        cache_path = os.path.normpath(cache_path)
        self.base_path = cache_path


class RedisCache(Cache):
    def __init__(self, name="redis", owner=None, *args, **kwargs):
        Cache.__init__(self, name=name, owner=owner, *args, **kwargs)

    def length(self):
        if self.hash:
            return self._redis.hlen(self.key)
        else:
            return len(self._redis.keys())

    def clear(self):
        if self.hash:
            self._redis.delete(self.key)
        else:
            self._redis.flushdb()

    def get_item(self, key):
        if self.hash:
            return self._get_item_hash(key)
        else:
            return self._get_item(key)

    def set_item(self, key, value, expires=None, timeout=None):
        if self.hash:
            return self._set_item_hash(key, value, expires=expires, timeout=timeout)
        else:
            return self._set_item(key, value, expires=expires, timeout=timeout)

    def delete_item(self, key):
        if self.hash:
            self._redis.hdel(self.key, key)
        else:
            self._redis.delete(key)

    def _load(self, *args, **kwargs):
        Cache._load(self, *args, **kwargs)
        self.hash = kwargs.pop("hash", False)
        self._redis = redisdb.get_connection()
        self._redis.ping()

    def _unload(self, *args, **kwargs):
        Cache._unload(self, *args, **kwargs)
        self._redis = None

    def _get_item(self, key):
        if not self._redis.exists(key):
            raise KeyError("not found")
        return self._redis.get(key)

    def _get_item_hash(self, key):
        if not self._redis.hexists(self.key, key):
            raise KeyError("not found")
        return self._redis.hget(self.key, key)

    def _set_item(self, key, value, expires=None, timeout=None):
        self._redis.set(key, value)
        if expires:
            timeout = expires - time.time()
        if timeout and timeout > 0:
            self._redis.expire(key, int(timeout))
        elif timeout:
            self._redis.delete(key)

    def _set_item_hash(self, key, value, expires=None, timeout=None):
        self._redis.hset(self.key, key, value)
        if expires:
            timeout = expires - time.time()
        if timeout and timeout > 0:
            self._redis.expire(self.key, int(timeout))
        elif timeout:
            self._redis.hdel(self.key, key)

    @property
    def key(self, prefix="cache"):
        suffix = self.owner.name_i if self.owner else "global"
        return prefix + ":" + suffix


class SerializedCache(object):
    SERIALIZER = pickle
    """ The serializer to be used for the values contained in
    the session (used on top of the class) """

    def __init__(self, cache, serializer=None):
        cls = self.__class__
        self._cache = cache
        self._serializer = serializer or cls.SERIALIZER

    def __len__(self):
        return self._cache.__len__()

    def __getitem__(self, key):
        return self.get_item(key)

    def __setitem__(self, key, value):
        return self.set_item(key, value)

    def __delitem__(self, key):
        return self._cache.__delitem__(key)

    def __contains__(self, item):
        return self._cache.__contains__(item)

    def __nonzero__(self):
        return self._cache.__nonzero__()

    def __bool__(self):
        return self._cache.__bool__()

    def __getattr__(self, name):
        if hasattr(self._cache, name):
            return getattr(self._cache, name)
        raise AttributeError("'%s' not found" % name)

    def get_item(self, key):
        data = self._cache.get_item(key)
        value = self._serializer.loads(data)
        return value

    def set_item(self, key, value, expires=None, timeout=None):
        data = self._serializer.dumps(value)
        return self._cache.set_item(key, data, expires=expires, timeout=timeout)
